#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
from pricelib.common.utilities.enums import CallPut, InOut, ExerciseType, EngineType, PaymentType, StatusType
from pricelib.common.time import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.patterns import Observer
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.fdm_engines import FdmBarrierEngine


class DoubleBarrierOption(OptionBase, Observer):
    """双边障碍期权
    (call或put，不是双鲨结构)"""

    def __init__(self, strike, callput: CallPut, inout: InOut, exercise_type=ExerciseType.American,
                 payment_type=PaymentType.Expire, *, bound=(80, 120), rebate=(0, 0), parti=1, engine=None,
                 status=StatusType.NoTouch, maturity=None, start_date=None, end_date=None, window=(None, None),
                 discrete_obs_interval=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365,
                 t_step_per_year=243, s=None, r=None, q=None, vol=None):
        """构造函数
        产品参数:
            strike: float, 执行价
            bound: (lower_barrier, upper_barrier), 低障碍价/高障碍价，障碍价格绝对数值
            callput: 看涨看跌类型，CallPut枚举类，Call/Put
            inout: 敲入敲出类型，InOut枚举类，In敲入/Out敲出
            exercise_type: 障碍观察类型，ExerciseType枚举类，美式观察American/欧式到期观察European
            payment_type: 现金返还支付类型，PaymentType枚举类
                          敲入现金返还一定是到期支付Expire，敲出可以是到期支付Expire或者障碍触发支付Barrier
            rebate: (lower_rebate, upper_rebate), 未敲出现金返还/敲入现金返还的绝对数值，非百分比，非年化
                                                  对敲出期权，元组第一位是低障碍价下方的现金返还，第二位是高障碍价上方的现金返还，
                                                  对敲入期权，元组第一位是敲入现金返还，第二位没有用处。
            parti: float, 香草期权的参与率，默认为1
            status: 敲入敲出状态，StatusType枚举类，默认为NoTouch未触碰，UpTouch为向上触碰，DownTouch为向下触碰
            engine: 定价引擎，PricingEngine类
                    解析解: AnalyticDoubleBarrierEngine 只支持美式观察(整个有效期观察)；支持连续观察/离散观察(默认为每日观察)；
                                                       没有现金返还，现金返还都是0
                    PDE: FdmBarrierEngine  只支持美式观察(整个有效期观察)；支持连续观察/离散观察(默认为每日观察)；
                                           支持现金返还；敲入现金返还为到期支付；敲出现金返还支持 立即支付/到期支付
                    蒙特卡洛: MCDoubleBarrierEngine  支持欧式观察(仅到期观察)/美式观察(整个有效期观察)；只支持离散观察(默认为每日观察)；
                                            支持现金返还；敲入现金返还为到期支付；敲出现金返还支持 立即支付/到期支付
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            window: (start_date, end_date)，障碍观察时间窗口，距离估值日的年化期限。 仅PDE定价引擎支持该参数。
                    如果是默认值None，说明观察区间是整个存续期: 估值日-终止日
            discrete_obs_interval: 观察时间间隔. 若为连续观察，None；若为均匀离散观察，为年化的观察时间间隔
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用PDE定价引擎 FdmBarrierEngine
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        self.strike = strike
        self.bound = bound  # (低障碍价, 高障碍价)
        self.rebate = rebate  # 双接触: (下边界payoff, 上边界payoff); 双不接触: (不接触payoff, 无作用)
        self.window = window  # (观察窗口起始时间, 观察窗口结束时间)
        self._callput = callput
        self._inout = inout
        self.exercise_type = exercise_type
        self.payment_type = payment_type
        self.parti = parti
        self.discrete_obs_interval = discrete_obs_interval  # 连续观察=None；均匀离散观察=观察时间间隔
        self.status = status
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            default_engine = FdmBarrierEngine(s=s, r=r, q=q, vol=vol)
            self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """设置定价引擎，同时将自己注册为观察者。若已有定价引擎，先将自己从原定价引擎的观察者列表中移除"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新是否已经触碰障碍"""
        if observable == self.engine.process.spot:
            if self.discrete_obs_interval is None:  # TODO: 目前update仅支持连续观察，离散观察需要判断是否是观察日
                if observable.value >= self.bound[1]:
                    self.status = StatusType.UpTouch
                elif observable.value <= self.bound[0]:
                    self.status = StatusType.DownTouch

    def __repr__(self):
        """返回期权的描述"""
        if self.discrete_obs_interval is None:
            obs_type = "连续观察"
        else:
            obs_type = "离散观察"
        return f"{obs_type}双边{self.inout.value}障碍{self.callput.name}期权"

    @property
    def callput(self):
        return self._callput

    @callput.setter
    def callput(self, value):
        self._callput = value
        self.status = StatusType.NoTouch  # 产品参数改变，重置是否已经敲出或敲入

    @property
    def inout(self):
        return self._inout

    @inout.setter
    def inout(self, value):
        self._inout = value
        self.status = StatusType.NoTouch  # 产品参数改变，重置是否已经敲出或敲入

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        if self.engine.engine_type == EngineType.PdeEngine:  # 接口特殊是因为PDE引擎兼容了单双边障碍
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot, bound=self.bound, rebate=self.rebate)
        else:
            price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0

日期处理主要功能实现：

1. Date: 日期类
datetime.datetime(year,month,day,hour=0,minute=0,second=0,microsecond=0,tzinfo=None,*,fold=0)
支持比较大小: ==, !=, >, <, >=, <=
支持获取年月日星期:year,month,day, weekday(), isoweekday()
支持加减时间: + timedelta(days=1) - relativedelta(weeks=1) + relativedelta(months=1) - relativedelta(years=1)

2. Period: 时间段类
2.1 datetime.timedelta(days=0, seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0)
主要使用timedelta的days，计算自然日时间差: (date1 - date2).days  若含头含尾，需+1

2.2 relativedelta: 处理周、月、年的时间差
    (self, dt1=None, dt2=None,
    years=0, months=0, days=0, leapdays=0, weeks=0,
    hours=0, minutes=0, seconds=0, microseconds=0,
    year=None, month=None, day=None, weekday=None,
    yearday=None, nlyearday=None,
    hour=None, minute=None, second=None, microsecond=None)
计算相对时间差: relativedelta(Date1, Date2)

3. Calendar: 交易日历类
自定义，详见下方代码

4. Schedule: 日期序列类
自定义，详见下方代码

todo: 日期规则，DayCounters
"""
from enum import Enum, unique
import datetime
from dateutil.relativedelta import relativedelta
from dateutil import rrule


class Calendar:
    """某个国家或地区的交易日历
        advance方法: new_date = calender.advance(old_date, period)
        addHoliday: 添加节假日
        removeHoliday: 移除节假日
        businessDaysBetween:获取交易日天数
    """

    def __init__(self):
        """初始化交易日历"""
        self.holidays = set()  # 节假日期集合

    def advance(self, date, period):
        """计算在当前交易日历下，指定日期加上一段时间后，生成的新日期
        Args:
            date: datetime.date，指定日期
            period: datetime.timedelta，时间段，单位是天
        Returns:
            new_date: datetime.date，新日期
        """
        if isinstance(date, datetime.datetime):
            date = date.date()
        new_date = date
        while period.days > 0:
            new_date += datetime.timedelta(days=1)
            if new_date.weekday() < 5 and new_date not in self.holidays:
                period -= datetime.timedelta(days=1)
        while period.days < 0:
            new_date -= datetime.timedelta(days=1)
            if new_date.weekday() < 5 and new_date not in self.holidays:
                period += datetime.timedelta(days=1)
        return new_date

    def business_list_between(self, dt1, dt2):
        """计算在当前交易日历下，dt1和dt2之间有哪些交易日(含头含尾)
        Args:
            dt1: datetime.datetime或datetime.date，起始日
            dt2: datetime.datetime或datetime.date，终止日
        Returns:
            business_days: List[datetime.date]，交易日列表(含头含尾)
        """
        if isinstance(dt1, datetime.datetime):
            dt1 = dt1.date()
        if isinstance(dt2, datetime.datetime):
            dt2 = dt2.date()
        days = rrule.rrule(freq=rrule.DAILY, byweekday=(rrule.MO, rrule.TU, rrule.WE, rrule.TH, rrule.FR),
                           dtstart=dt1, until=dt2)  # rrule生成的是datetime.datetime对象
        business_days = [day.date() for day in days if day.date() not in self.holidays]  # 0-4代表周一到周五
        return business_days

    def business_days_between(self, dt1, dt2):
        """计算在当前交易日历下，dt1和dt2之间有多少个交易日(不含头，含尾)
        Args:
            dt1: datetime.datetime或datetime.date，起始日
            dt2: datetime.datetime或datetime.date，终止日
        Returns:
            int，交易日天数(不含头，含尾)
        """
        return len(self.business_list_between(dt1, dt2)) - 1

    def add_holiday(self, date):
        """添加节假日
        Args:
            date: datetime.date，节假日
        Returns: None
        """
        self.holidays.add(date)

    def remove_holiday(self, date):
        """移除非节假日
        Args:
            date: datetime.date，非节假日
        Returns: None
        """
        self.holidays.remove(date)


@unique
class BusinessConvention(Enum):
    """枚举类，用于指定日期调整规则
    例如Scheduler生成敲出观察日时，遇到非交易日，需要提前/延后一个工作日"""
    PRECEDING = "提前"
    FOLLOWING = "延后"


@unique
class MonthAdjustment(Enum):
    """遇到非交易日调整时，如果跨月，是否调整为不换月"""
    YES = 0
    NO = 1


@unique
class EndOfMonthModify(Enum):
    """到期日是非交易日，调整到期日时，恰逢起始日是某月最后一个交易日，终止日是非交易日，向前和向后调整终止日后月份不同，选择向前调整"""
    YES = 0
    NO = 1


# 每周 - 7日，每双周 - 14日，每月 - 30.416日，每季度 - 91.25日
FREQS = {'w': 7, 'd': 14, 'm': 30.416, 's': 91.25}
FREQS_RELATIVEDELTA = {'w': relativedelta(weeks=1), 'd': relativedelta(weeks=2),
                       'm': relativedelta(months=1), 's': relativedelta(months=3)}


class Schedule:
    """日期序列类
    根据起止日期、观察频率和交易日规则，生成观察日列表
    根据某个基准日期，计算self.date_schedule中的每个日期距离基准日期几个自然日(交易日)，返回一个List[int]，列表中是自然日(交易日)数量
    """

    def __init__(self, trade_calendar, start=None, end=None, freq='m', lock_term=1,
                 convention=BusinessConvention.FOLLOWING, correction=MonthAdjustment.YES,
                 endofmonthmodify=EndOfMonthModify.YES, date_schedule=None):
        """初始化Schedule
        Args:
            trade_calendar: 某个市场的Calendar交易日历对象
            start: datetime.date，起始日期
            end: datetime.date，终止日期
            freq: str，观察频率，'w' - 每周，'d' - 每双周，'m' - 每月，'s' - 每季度
            lock_term: int，锁定期，单位是观察频率，比较常见的是敲出观察锁3月
            convention: 日期调整规则，BusinessConvention枚举类，'提前'或'延后'
            correction: 非交易日跨月调整规则，MonthAdjustment枚举类，YES - 调整为不换月，NO - 不调整
            endofmonthmodify: 到期日跨月调整规则，EndOfMonthModify枚举类，YES - 向前调整，NO - 不调整
            date_schedule: List[datetime.date]，日期列表，可以直接外部输入
                           默认为None，会根据start, end, freq, lock_term, convention, correction, endofmonthmodify生成
        """
        self.start = start
        self.end = end
        self.freq = freq
        self.lock_term = lock_term
        self.convention = convention
        self.correction = correction
        self.endofmonthmodify = endofmonthmodify
        self.trade_calendar = trade_calendar
        args = (start, end, freq, lock_term, convention, correction, endofmonthmodify)
        if date_schedule is None:  # 如果没有直接输入日期列表，检查参数是否有None，生成date_schedule
            assert any(arg is not None for arg in args), "Error: Schedule输入参数不全，有None"
            self.date_schedule = self.generate_schedule()
        else:  # 直接输入日期列表
            self.date_schedule = date_schedule
        self.count_business, self.count_calendar = None, None

    def generate_schedule(self):
        """根据起止日期、频率和交易日规则，生成日期列表"""
        # 如果到期日非交易日，先按照月终规则调整到期日：如果起始日是某月最后一个交易日，终止日是非交易日，向前和向后调整终止日后月份不同，选择向前调整
        if (self.endofmonthmodify == EndOfMonthModify.YES and (
                self.end in self.trade_calendar.holidays or self.end.weekday() > 4)):
            next_startday = self.trade_calendar.advance(self.start, datetime.timedelta(days=1))
            # 如果起始日是该月最后一个交易日
            if self.start.month != next_startday.month:
                prev_endday = self.trade_calendar.advance(self.end, datetime.timedelta(days=-1))
                next_endday = self.trade_calendar.advance(self.end, datetime.timedelta(days=1))
                # 如果到期日向前调整和向后调整跨月，向前调整
                if prev_endday.month != next_endday.month:
                    self.end = prev_endday
        # 起始日（含）和终止日（含）之间所有自然日期，统计总日数N
        n_dates = (self.end - self.start).days + 1
        # 根据敲出观察频率确定敲出观察日之间间隔日数f：每月 - 30日，每季度 - 90，每周 - 7，每双周 - 14
        fq = FREQS[self.freq]
        # 根据间隔日数计算敲出观察次数
        n_obs = round(n_dates / fq) + 1
        # 根据观察次数和总日数，使用relativedelta每次加上一个间隔，推算各个日期
        n_obs_dates = [self.start + i * FREQS_RELATIVEDELTA[self.freq] for i in range(n_obs)][self.lock_term:]
        # 根据观察开始时间i，筛选观察日序号
        self.date_schedule = []
        for i, obs_date in enumerate(n_obs_dates):
            if obs_date in self.trade_calendar.holidays or obs_date.weekday() > 4:  # 如果 obs_date不是交易日
                if self.convention == BusinessConvention.PRECEDING:  # 提前
                    obs_date_adj = self.trade_calendar.advance(obs_date, datetime.timedelta(days=-1))
                    if self.correction == MonthAdjustment.YES and obs_date_adj.month != obs_date.month:
                        # 如果提前导致换月，改为延后
                        obs_date_adj = self.trade_calendar.advance(obs_date, datetime.timedelta(days=1))
                else:  # 延后
                    obs_date_adj = self.trade_calendar.advance(obs_date, datetime.timedelta(days=1))
                    if self.correction == MonthAdjustment.YES and obs_date_adj.month != obs_date.month:
                        # 如果延后导致换月，改为提前
                        obs_date_adj = self.trade_calendar.advance(obs_date, datetime.timedelta(days=-1))
                self.date_schedule.append(obs_date_adj)
            else:
                self.date_schedule.append(obs_date)
        # 若最后一个观察日为非交易日，将其延后之后，如果最后一个观察日超过终止日，最后一个观察日改为提前
        while self.date_schedule[-1] > self.end:
            self.date_schedule[-1] = self.trade_calendar.advance(self.date_schedule[-1], datetime.timedelta(days=-1))
        # 使用 dict.fromkeys() 去重，并保持顺序
        self.date_schedule = list(dict.fromkeys(self.date_schedule))
        return self.date_schedule

    def count_business_days(self, base_date):
        """根据某个基准日期，计算self.date_schedule中的每个日期距离基准日期几个交易日
           返回一个列表，列表中是交易日数量int
        Args:
            base_date: datetime.date，基准日期
        Returns:
            self.count_business: List[int]，交易日数量列表
        """
        self.count_business = [self.trade_calendar.business_days_between(base_date, date)
                               for date in self.date_schedule]
        return self.count_business

    def count_calendar_days(self, base_date):
        """根据某个基准日期，计算self.date_schedule中的每个日期距离基准日期几个自然日，
           返回一个列表，列表中是自然日数量int
        Args:
            base_date: datetime.date，基准日期
        Returns:
            self.count_calendar: List[int]，自然日数量列表
        """
        self.count_calendar = [(date - base_date).days for date in self.date_schedule]
        return self.count_calendar


@unique
class AnnualDays(Enum):
    """年化系数枚举类"""
    N365 = 365
    N360 = 360
    N243 = 243
    N244 = 244



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import abstractmethod
import numpy as np
from scipy.interpolate import interp1d
from pricelib.common.utilities.enums import StatusType, ExerciseType
from pricelib.common.pricing_engine_base import FdmEngine, FdmGrid
from pricelib.common.time import global_evaluation_date


class FdmAutoCallableEngine(FdmEngine):
    """AutoCallable PDE有限差分法定价引擎基类
        雪球/凤凰/FCN/DCN, 支持变敲出、变敲入、变票息等要素可变型结构"""

    def __init__(self, stoch_process=None, s_step=800, n_smax=2, fdm_theta=1, *,
                 s=None, r=None, q=None, vol=None):
        """初始化有限差分法定价引擎
        Args:
            stoch_process: StochProcessBase随机过程
            s_step: int，价格步数
            n_smax: int，价格网格上界，设为n_smax倍初始价格
            fdm_theta: float，时间方向有限差分theta，0：explicit, 1: implicit, 0.5: Crank-Nicolson
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process, s_step=s_step, n_smax=n_smax, fdm_theta=fdm_theta, s=s, r=r, q=q, vol=vol)
        # 计算过程的中间变量
        self.prod = None  # Product产品对象
        self.out_dates = None  # 根据估值日，将敲出观察日转化为List[int]，交易日期限
        self.pay_dates = None  # 根据估值日，将支付日转化为List[float]，年化自然日期限
        self.diff_obs_pay_dates = None  # 观察日与付息日的差值，List[float]，年化自然日期限
        self.smax = None  # 价格网格上界，float
        self.t_step = None  # 网格的时间步数，int
        self.dt = None  # 时间步长，float
        self.j_vec = None  # 时间格点，List[int]
        self.fd_not_in = None  # 未敲入的有限差分网格FdmGrid对象
        self.fd_knockin = None  # 已敲入的有限差分网格FdmGrid对象
        self._barrier_in = None  # 是经过估值日截断之后的列表，例如prod.barrier_in有22个，存续一年时估值，self._barrier_in只有12个
        self._barrier_out = None  # 是经过估值日截断之后的列表，例如prod.barrier_out有22个，存续一年时估值，self._barrier_out只有12个
        self.next_paydate = None
        self.next_barrier_in = None
        self.reversed_barrier_in = None
        self.next_barrier_out = None
        self.reversed_barrier_out = None
        self.in_idx = None
        self.out_idxs = None
        # 雪球独有
        self.next_coupon_out = None
        self.reversed_coupon_out = None
        self._coupon_out = None  # 是经过估值日截断之后的列表，例如prod.coupon_out有22个，存续一年时估值，self._coupon_out只有12个
        self.coupon_outs = None
        self.itm = None
        self.next_diff_obspaydate = None
        # 凤凰独有
        self._barrier_yield = None  # 是经过估值日截断之后的列表，例如prod.barrier_yield有24个，存续一年时估值，self._barrier_yield只有12个
        self.next_barrier_yield = None
        self.reversed_barrier_yield = None
        self.next_coupon = None
        self._coupon = None  # 是经过估值日截断之后的列表，例如prod.coupon有24个，存续一年时估值，self._coupon只有12个
        self.coupons = None
        self.reversed_coupon = None
        self.yld_idxs = None

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date,
                                                                            prod.end_date) / prod.t_step_per_year
        obs_dates = prod.obs_dates.count_business_days(calculate_date)
        obs_dates = np.array([num for num in obs_dates if num >= 0])
        calculate_start_diff = (calculate_date - prod.start_date).days
        pay_dates = prod.pay_dates.count_calendar_days(prod.start_date)
        self.pay_dates = np.array([num / prod.annual_days.value for num in pay_dates if num >= calculate_start_diff])
        assert len(obs_dates) == len(self.pay_dates), f"Error: {prod}的观察日和付息日长度不一致"
        self.diff_obs_pay_dates = np.array([(prod.pay_dates.date_schedule[i] - prod.obs_dates.date_schedule[i]).days
                                            for i in range(len(obs_dates))]) / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()

        self.smax = self.n_smax * prod.s0  # 价格网格上界, 默认为n_smax倍初始价格

        # 返回PDE系数组件abc的函数
        fn_pde_coef = self.process.get_fn_pde_coef(_maturity, spot)

        # 统一已敲入、未敲入矩阵的格点
        self.t_step = round(prod.t_step_per_year * _maturity_business_days)  # 时间步数
        self.dt = _maturity_business_days / self.t_step  # 时间步长
        self.j_vec = np.linspace(0, self.t_step, self.t_step + 1)  # 生成时间格点

        # 初始化已敲入、未敲入的FdmGrid对象
        self.fd_not_in = FdmGrid(smax=self.smax, maturity=_maturity_business_days,
                                 t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                                 fn_pde_coef=fn_pde_coef, fdm_theta=self.fdm_theta)  # 未敲入, [0, smax]
        self.fd_knockin = FdmGrid(smax=self.smax, maturity=_maturity_business_days,
                                  t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                                  fn_pde_coef=fn_pde_coef, fdm_theta=self.fdm_theta)  # 已敲入，[0, smax]

        # 初始化逆序敲出观察日
        if isinstance(obs_dates, (list, np.ndarray)):  # 敲出观察日是具体第n天数的列表，如[11,22,32,43,53,64]
            out_dates = np.round(np.flip(np.array(obs_dates))).astype(int)
        elif isinstance(obs_dates, (int, np.int32)):  # 敲出观察日是整数， 敲出观察日个数，如6代表6个均匀分布的观察日
            out_dates = (self.t_step - np.round(np.linspace(0, self.t_step, round(obs_dates + 1)))[:-1])
            out_dates = out_dates.astype('int')
        else:
            raise ValueError('敲出观察日类型设置错误，请检查')
        self.out_dates = out_dates

        # 初始化到期日payoff价值
        self._init_terminal_condition(self.fd_not_in.s_vec, _maturity)

        # 初始化已敲入、未敲入的边界条件
        self._init_boundary_condition(self.smax, _maturity)

        # N为时间步数，从倒数第二天向第一天反向递推
        self._backward_induction()

        if prod.status == StatusType.DownTouch:
            result = self.fd_knockin.functionize(self.fd_knockin.v_grid[1:-1, 0], kind="cubic")(spot)
        else:
            result = self.fd_not_in.functionize(self.fd_not_in.v_grid[1:-1, 0], kind="cubic")(spot)
        return np.float64(result)

    @abstractmethod
    def _init_terminal_condition(self, *args, **kwargs):
        """初始化终止条件"""

    @abstractmethod
    def _init_boundary_condition(self, *args, **kwargs):
        """初始化边界条件"""

    @abstractmethod
    def _backward_induction(self, *args, **kwargs):
        """反向递推"""

    def delta(self, prod, t=0, spot=None, step=None, status: StatusType = None):
        """求t时刻，价格spot的delta值,
            spot是价格的绝对值
            t是期权价值矩阵的列(时间)的索引
            step为微扰步长，默认为价格spot的1.0%，
            status为敲入状态"""
        spot = self.process.spot() if spot is None else spot
        step = spot * 0.01 if step is None else step
        status = prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        f = fdm_grid.functionize(fdm_grid.v_grid[1:-1, t], kind='cubic')
        delta = (f(min(spot + step, self.smax)) - f(min(spot - step, self.smax))) / (2 * step)
        return delta

    def gamma(self, prod, t=0, spot=None, step=None, status: StatusType = None):
        """求t时刻，价格spot的gamma值,
            spot是价格的绝对值
            t是期权价值矩阵的列(时间)的索引
            step为微扰步长，默认为价格spot的1.0%，
            status为敲入状态"""
        spot = self.process.spot() if spot is None else spot
        step = spot * 0.01 if step is None else step
        status = prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        f = interp1d(fdm_grid.s_vec, fdm_grid.v_grid[1:-1, t], kind='cubic')
        gamma = (f(spot + step) - 2 * f(spot) + f(spot - step)) / (step ** 2)
        return gamma

    def theta(self, prod, t=0, spot=None, step=1, status: StatusType = None):
        '''默认: 一天的theta'''
        spot = self.process.spot() if spot is None else spot
        status = prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        f = fdm_grid.functionize(fdm_grid.v_grid[1:-1, t], kind='cubic')
        g = fdm_grid.functionize(fdm_grid.v_grid[1:-1, t + step], kind='cubic')
        theta = (g(spot) - f(spot)) / (step * spot)  # ( step * self.dt)
        return theta

    def vega(self, prod, t=0, spot=None, step=0.01, status: StatusType = None):
        """vega todo:目前只支持常数波动率"""
        spot = self.process.spot() if spot is None else spot
        status = prod.status if status is None else status
        last_vol = self.process.vol.volval
        last_in_grid = self.fd_knockin.v_grid.copy()
        last_not_grid = self.fd_not_in.v_grid.copy()
        fdm_grid0 = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        f = fdm_grid0.functionize(fdm_grid0.v_grid[1:-1, t], kind='cubic')
        self.process.vol.volval = last_vol + step
        self.calc_present_value(prod=prod)
        fdm_grid1 = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        g = fdm_grid1.functionize(fdm_grid1.v_grid[1:-1, t], kind='cubic')
        vega = (g(spot) - f(spot)) / spot
        self.process.vol.volval = last_vol
        self.fd_knockin.v_grid = last_in_grid
        self.fd_not_in.v_grid = last_not_grid
        return vega

    def rho(self, prod, t=0, spot=None, step=0.01, status: StatusType = None):
        """rho todo:目前只支持常数无风险利率"""
        spot = self.process.spot() if spot is None else spot
        status = prod.status if status is None else status
        last_r = self.process.interest.data
        last_in_grid = self.fd_knockin.v_grid.copy()
        last_not_grid = self.fd_not_in.v_grid.copy()
        fdm_grid0 = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        f = fdm_grid0.functionize(fdm_grid0.v_grid[1:-1, t], kind='cubic')
        self.process.interest.data = last_r + step
        self.calc_present_value(prod=prod)
        fdm_grid1 = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        g = fdm_grid1.functionize(fdm_grid1.v_grid[1:-1, t], kind='cubic')
        rho = (g(spot) - f(spot)) / spot
        self.process.interest.data = last_r
        self.fd_knockin.v_grid = last_in_grid
        self.fd_not_in.v_grid = last_not_grid
        return rho

    # pylint: disable=too-many-locals
    def pv_and_greeks(self, prod, t=0, spot=None, status: StatusType = None):
        """当一次性计算pv和5个greeks时，可以调用此函数
        Returns:
            Dict[str:float]: {'pv': pv, 'delta': delta, 'gamma': gamma, 'theta': theta, 'vega': vega, 'rho': rho}"""
        spot = self.process.spot() if spot is None else spot
        status = prod.status if status is None else status
        s_step = spot * 0.01
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        f = fdm_grid.functionize(fdm_grid.v_grid[1:-1, t], kind='cubic')
        pv = np.float64(f(spot))
        delta = (f(min(spot + s_step, self.smax)) - f(min(spot - s_step, self.smax))) / (2 * s_step)
        gamma = (f(spot + s_step) - 2 * f(spot) + f(spot - s_step)) / (s_step ** 2)

        g = fdm_grid.functionize(fdm_grid.v_grid[1:-1, t + 1], kind='cubic')
        theta = (g(spot) - f(spot)) / (1 * spot)

        last_in_grid = self.fd_knockin.v_grid.copy()
        last_not_grid = self.fd_not_in.v_grid.copy()

        last_vol = self.process.vol.volval
        self.process.vol.volval = last_vol + 0.01
        self.calc_present_value(prod=prod)
        fdm_grid1 = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        g1 = fdm_grid1.functionize(fdm_grid1.v_grid[1:-1, t], kind='cubic')
        vega = (g1(spot) - f(spot)) / spot
        self.process.vol.volval = last_vol

        last_r = self.process.interest.data
        self.process.interest.data = last_r + 0.01
        self.calc_present_value(prod=prod)
        fdm_grid2 = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        g2 = fdm_grid2.functionize(fdm_grid2.v_grid[1:-1, t], kind='cubic')
        rho = (g2(spot) - f(spot)) / spot
        self.process.interest.data = last_r

        self.fd_knockin.v_grid = last_in_grid
        self.fd_not_in.v_grid = last_not_grid

        return {'pv': pv, 'delta': delta, 'gamma': gamma, 'vega': vega, 'theta': theta, 'rho': rho}

    def value_matrix(self, status: StatusType = None):
        """根据是否敲入状态，返回S-t矩阵"""
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        return fdm_grid.v_grid, fdm_grid.s_vec

    def delta_matrix(self, step=0, status: StatusType = None):
        """计算S-t矩阵中，每个t时刻所有s格点的delta
            step: 标的价步长，默认为0，即选择PDE有限差分网格的价格步长
            status为敲入状态"""
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        if step == 0:
            # 默认PDE模拟时的步长 step = fdm_grid.ds
            delta_s_t = (fdm_grid.v_grid[2:] - fdm_grid.v_grid[:-2]) / (2 * fdm_grid.ds)
            return delta_s_t, fdm_grid.s_vec  # 中心差分，价格向量掐头0去尾smax，正好是fdm_grid.s_vec
        else:
            # 重新定义步长，在模拟矩阵的基础上插值
            n_step = np.round(self.smax / step).astype('int')
            spots = np.arange(0, n_step, 1) * step
            # 下方n_step和spots用于避免计算远离敲入水平的场景，降低运算量
            # n_step = 2 * np.round((self.spot - self.barrier_in + 1) / step)
            # spots = np.arange(0, n_step, 1) * step + self.barrier_in - 1
            delta_s_t = np.zeros((n_step - 1, fdm_grid.v_grid.shape[1]))
            # 分别在fdm_grid.s_vec价格向量首位添加0和smax
            x_vec = np.insert(fdm_grid.s_vec, 0, 0)
            x_vec = np.append(x_vec, self.smax)
            for j in range(fdm_grid.v_grid.shape[1]):
                f = interp1d(x_vec, fdm_grid.v_grid[:, j], kind='cubic')
                for i, spot in enumerate(spots[1:]):
                    delta_s_t[i, j] = (f(spot + step) - f(spot - step)) / (step * 2)
            return delta_s_t, spots

    def gamma_matrix(self, step=0, status: StatusType = None):
        """计算S-t矩阵中，每个t时刻所有s格点的gamma
            step: 标的价步长，默认为0，即选择PDE有限差分网格的价格步长
            status为敲入状态"""
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        if step == 0:
            # 默认PDE模拟时的步长 step = fdm_grid.ds
            gamma_s_t = np.diff(fdm_grid.v_grid, n=2, axis=0) / fdm_grid.ds ** 2
            return gamma_s_t, fdm_grid.s_vec
        else:
            # 重新定义步长，在模拟矩阵的基础上插值
            n_step = np.round(self.smax / step).astype('int')
            step = self.smax / n_step
            spots = np.arange(0, n_step, 1) * step

            gamma_s_t = np.zeros((n_step - 2, fdm_grid.v_grid.shape[1]))
            # 分别在fdm_grid.s_vec价格向量首位添加0和smax
            x_vec = np.insert(fdm_grid.s_vec, 0, 0)
            x_vec = np.append(x_vec, self.smax)
            for j in range(fdm_grid.v_grid.shape[1]):
                f = interp1d(x_vec, fdm_grid.v_grid[:, j], kind='cubic')
                for i, spot in enumerate(spots[1:-2]):
                    gamma_s_t[i, j] = (f(spot + step) - 2 * f(spot) + f(spot - step)) / (step ** 2)
            return gamma_s_t, spots

    def theta_matrix(self, step=0, status: StatusType = None):
        """计算S-t矩阵中，每个s价格所有t时刻的theta'
            step: 时间步长，默认为0，即选择PDE有限差分网格的时间步长
            status为敲入状态"""
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        if step == 0:
            # 默认PDE模拟时的步长
            spot = self.process.spot()
            theta_s_t = np.diff(fdm_grid.v_grid, axis=1) / spot  # / self.dt
            x_vec = np.insert(fdm_grid.s_vec, 0, 0)
            x_vec = np.append(x_vec, self.smax)
            return theta_s_t, x_vec
        else:
            raise NotImplementedError("todo: 时间步长与pde有限差分网格不同的theta.")


class FdmSnowBallEngine(FdmAutoCallableEngine):
    """ 雪球类 AutoCallable PDE 有限差分法定价引擎
        自动赎回Autocall Note(无敲入)/雪球式Snowball Autocallable/触发器Trigger Autocallable
            自动赎回：无敲入（敲入价为-1），有红利票息（未敲出有保底收益），为保本产品
            雪球式：有敲入，敲入后下跌部分保护，敲出年化票息
            触发器：有敲入，敲入后下跌部分保护，敲出绝对票息
        支持变敲出、变票息、变敲入、雪球plus、保底雪球、OTM、触发式等多种结构
    """

    def _init_boundary_condition(self, smax, maturity):
        """初始化边界条件
        考虑到了保底、OTM、保证金/纯期权模式、以及敲出边界上方增加虚值call(增强型)
        Args:
            smax: float, 标的价格上界
            maturity: float, 到期时间，年化自然日期限
        Returns: None
        """
        prod = self.prod
        # 矩阵第一行: 标的价格为0时结构的价值
        t_vec = self.dt * self.j_vec
        interest_maturity_discount_factor = self.process.interest.disc_factor(maturity, t_vec)

        # 由看涨看跌平价 call-put=S*exp(-qt)-K*exp(-rt)，S=0时call=0，put=K*exp(-rt)
        self.fd_not_in.v_grid[0, :] = self.fd_knockin.v_grid[0, :] = (
                (prod.margin_lvl * prod.s0 - prod.strike_upper + prod.strike_lower)
                * interest_maturity_discount_factor)

        # 矩阵最后一行: 每一个观察日前，股价上界能得到的本金加票息贴现
        self.next_paydate = self.pay_dates.repeat(
            np.diff(np.append(np.zeros((1,)), self.out_dates[::-1])).astype(int))  # 计息时间数
        self.next_paydate = np.append(self.pay_dates[0], self.next_paydate)  # 每个交易日对应的下一个敲出观察日对应的付息日

        # 触发器：上边界为绝对票息贴现/ 非触发器：上边界为年化票息贴现
        coupon_t = 1 if prod.trigger else self.next_paydate
        # 由看涨看跌平价 call-put = S*exp(-qt) - K*exp(-rt)，S=Smax时put=0，call = Smax*exp(-qt) - K*exp(-rt)
        interest_next_paydate_discount_factor = self.process.interest.disc_factor(self.next_paydate, t_vec)
        div_next_paydate_discount_factor = self.process.div.disc_factor(self.next_paydate, t_vec)
        v_smax = ((smax * div_next_paydate_discount_factor
                   - prod.strike_call * interest_next_paydate_discount_factor)
                  * prod.parti_out + (prod.margin_lvl + self.coupon_outs * coupon_t) * prod.s0
                  * interest_next_paydate_discount_factor)
        self.fd_knockin.v_grid[self.s_step, :] = self.fd_not_in.v_grid[self.s_step, :] = v_smax

    def _init_terminal_condition(self, s_vec, maturity):
        """初始化终止时已敲入且未敲出、未敲入且未敲出的期权价值
        Args:
            s_vec: List[float], 网格的价格格点
            maturity: float, 到期时间，年化自然日期限
        Returns: None
        """
        prod = self.prod
        # 初始化next_barrier_in、next_barrier_out、next_coupon_out为最后一期的敲入价、敲出价、敲出票息。若变敲出变敲入，需要动态调整
        # 设置到期日的敲入边界位置
        if isinstance(prod.barrier_in, (int, float, np.int32, np.float64)):
            # 常数敲入雪球
            self.next_barrier_in = prod.barrier_in
        elif isinstance(prod.barrier_in, (list, np.ndarray)):
            # self.out_dates是估值日之后的、经过截断的敲出观察日，取产品参数列表中最后len(self.out_dates)个，与其对应
            self._barrier_in = prod.barrier_in[-len(self.out_dates):].copy()
            # 浮动敲入雪球
            self.next_barrier_in = self._barrier_in[-1]
            self.reversed_barrier_in = np.flip(np.array(self._barrier_in))
        else:
            raise ValueError(f'敲入障碍价类型为{type(prod.barrier_in)}，仅支持int/float/list/np.ndarray，请检查')

        # 设置到期日的敲出边界位置
        if isinstance(prod.barrier_out, (int, float, np.int32, np.float64)):
            # 常数敲出雪球
            self.next_barrier_out = prod.barrier_out
        elif isinstance(prod.barrier_out, (list, np.ndarray)):
            # self.out_dates是估值日之后的、经过截断的敲出观察日，取产品参数列表中最后len(self.out_dates)个，与其对应
            self._barrier_out = prod.barrier_out[-len(self.out_dates):].copy()
            # 阶梯式雪球
            self.next_barrier_out = self._barrier_out[-1]
            self.reversed_barrier_out = np.flip(np.array(self._barrier_out))
        else:
            raise ValueError(f'敲出障碍价类型为{type(prod.barrier_out)}，仅支持int/float/list/np.ndarray，请检查')

        # 敲出票息
        if isinstance(prod.coupon_out, (int, float, np.int32, np.float64)):
            # 常数敲出票息
            self.next_coupon_out = prod.coupon_out
            self.coupon_outs = np.full(self.t_step + 1, prod.coupon_out)  # N天的敲出票息列表，用于生成mspot价格上边界条件
        elif isinstance(prod.coupon_out, (list, np.ndarray)):
            # self.out_dates是估值日之后的、经过截断的敲出观察日，取产品参数列表中最后len(self.out_dates)个，与其对应
            self._coupon_out = prod.coupon_out[-len(self.out_dates):].copy()
            # 浮动敲出票息
            self.next_coupon_out = self._coupon_out[-1]
            self.reversed_coupon_out = np.flip(np.array(self._coupon_out))
            # N天的敲出票息列表，用于生成smax价格上边界条件
            coupon_outs = np.array(self._coupon_out).repeat(
                np.diff(np.append(np.zeros((1,)), self.out_dates[::-1])).astype(int))
            self.coupon_outs = np.append(coupon_outs[0], coupon_outs)
        else:
            raise ValueError(f'敲出票息类型为{type(prod.coupon_out)}，仅支持int/float/list/np.ndarray，请检查')

        # in_idx：小于到期日敲入线的最大的标的价格索引。如果是变敲入雪球，此变量in_idx应该随敲入价动态调整。
        self.in_idx = 1 + np.where(s_vec <= self.next_barrier_in)[0][-1] if self.next_barrier_in > 0 else 0
        # out_idxs：大于到期日敲出线的标的价格数组的索引。如果是变敲出雪球，此变量out_idxs应该随敲出价动态调整。
        self.out_idxs = 1 + np.array(np.where(s_vec >= self.next_barrier_out)[0])
        # 虚值call在值度
        self.itm = np.where(s_vec - prod.strike_call > 0, (s_vec - prod.strike_call) * prod.parti_out, 0)

        coupon_t = 1 if prod.trigger else self.pay_dates[-1]  # 票息是否年化
        # 矩阵最后一列：到期payoff。考虑到了保底、保证金\纯期权模式、以及敲出边界上方增加虚值call
        """已敲入：如果到期小于敲出边界，在100%本金模式下，小于保底边界，获得保底价；
                                                  大于保底边界，获得到期价格；
                                   在保证金模式下，小于保底边界，支付期初价（高行权价）与保底价之间价差；
                                                  大于保底边界，支付到期价格与期初价之间价差；
                  如果到期大于敲出边界，大于看涨行权价，获得看涨收益、本金和票息；
                                    在敲出边界和看涨行权价之间，获得本金和票息"""
        self.fd_knockin.v_grid[1:-1, self.t_step] = np.where(s_vec < self.next_barrier_out,
                                                             np.minimum(
                                                                 np.where(s_vec <= prod.strike_lower, prod.strike_lower,
                                                                          s_vec) -
                                                                 prod.strike_upper, 0) + prod.s0 * prod.margin_lvl,
                                                             np.where(s_vec > prod.strike_call,
                                                                      (s_vec - prod.strike_call) * prod.parti_out + (
                                                                              prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0,
                                                                      (
                                                                              prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0))
        """未敲入：如果到期小于敲入边界，在100%本金模式下，小于保底边界，获得保底价；
                                                  大于保底边界，获得实际价格；
                                  在保证金模式下，小于保底边界，支付保底价与期初价之间价差；
                                                  大于保底边界，支付到期价格与期初价之间价差；
                 如果到期大于敲入边界，且大于看涨行权价，获得看涨收益、本金和票息；
                                  在敲入边界和看涨行权价之间，获得本金和票息
                                          如果在敲出边界和看涨行权价之间，获得本金和敲出票息
                                          如果在敲入边界和敲出边界之间，获得本金和红利票息"""
        self.fd_not_in.v_grid[1:-1, self.t_step] = np.where(s_vec <= self.next_barrier_in,
                                                            np.where(s_vec <= prod.strike_lower, prod.strike_lower,
                                                                     s_vec) -
                                                            prod.strike_upper + prod.s0 * prod.margin_lvl,
                                                            np.where(s_vec > prod.strike_call,
                                                                     (s_vec - prod.strike_call) * prod.parti_out + (
                                                                             prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0,
                                                                     np.where(s_vec >= self.next_barrier_out,
                                                                              (
                                                                                      prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0,
                                                                              (
                                                                                      prod.margin_lvl + prod.coupon_div * coupon_t) * prod.s0)))

    def _backward_induction(self):
        """从倒数第二天向第一天逆向迭代"""
        prod = self.prod
        self.next_diff_obspaydate = self.diff_obs_pay_dates.repeat(
            np.diff(np.append(np.zeros((1,)), self.out_dates[::-1])).astype(int))  # 计息时间数
        self.next_diff_obspaydate = np.append(self.next_diff_obspaydate[0],
                                              self.next_diff_obspaydate)  # 每个交易日对应的下一个敲出观察日对应的付息日与观察日的自然日年化之差

        if prod.in_obs_type == ExerciseType.European:  # 敲入观察为欧式，仅到期观察敲入
            self.next_barrier_in = -1
            self.in_idx = 0
        for j in range(self.t_step - 1, -1, -1):
            self.fd_knockin.v_grid[1:-1, j] = self.fd_knockin.evolve(j, self.fd_knockin.v_grid[1:-1, j + 1], self.dt)
            self.fd_not_in.v_grid[1:-1, j] = self.fd_not_in.evolve(j, self.fd_not_in.v_grid[1:-1, j + 1], self.dt)

            # 考虑敲出的情况: 敲出观察日边界条件需更改
            if j in self.out_dates:
                coupon_t = 1 if prod.trigger else self.next_paydate[j]
                # 如果敲出价是浮动的，调整敲出价
                self.next_barrier_out = (self.reversed_barrier_out[np.where(self.out_dates == j)[0][0]]
                                         if isinstance(prod.barrier_out, (list, np.ndarray)) else prod.barrier_out)
                self.out_idxs = 1 + np.array(np.where(self.fd_not_in.s_vec >= self.next_barrier_out)[0])
                # 如果敲出票息是浮动的，调整敲出票息
                self.next_coupon_out = (self.reversed_coupon_out[np.where(self.out_dates == j)[0][0]]
                                        if isinstance(prod.coupon_out, (list, np.ndarray)) else prod.coupon_out)
                # 发生敲出的payoff
                knock_out_payoff = (self.itm[self.out_idxs - 1] + (
                        prod.margin_lvl + self.next_coupon_out * coupon_t) * prod.s0
                                    ) * self.process.interest.disc_factor(self.next_paydate[j], self.next_paydate[j] -
                                                                          self.next_diff_obspaydate[j])
                # 用knock_out_payoff覆盖敲出价上方的衍生品价值
                self.fd_not_in.v_grid[self.out_idxs, j] = self.fd_knockin.v_grid[self.out_idxs, j] = knock_out_payoff

                if prod.in_obs_type != ExerciseType.European:
                    # 如果敲入价是浮动的，调整敲入价
                    self.next_barrier_in = self.reversed_barrier_in[np.where(self.out_dates == j)[0][0]] if isinstance(
                        prod.barrier_in, (list, np.ndarray)) else prod.barrier_in
                    self.in_idx = 1 + np.where(self.fd_not_in.s_vec <= self.next_barrier_in)[0][
                        -1] if self.next_barrier_in > 0 else 0

            # 考虑敲入的情况: 向下越过敲入边界时，下一时间点敲入边界以上部分是标准autocall价格
            self.fd_not_in.v_grid[:(self.in_idx + 1), j] = self.fd_knockin.v_grid[:(self.in_idx + 1), j]


class FdmPhoenixEngine(FdmAutoCallableEngine):
    """ 凤凰类 AutoCallable PDE 有限差分法定价引擎：
            较雪球式AutoCallable，增加派息价格，少了敲出和红利票息
            每个派息（敲出）观察日，如果价格高于派息价格，派固定利息，如果发生敲出，合约提前结束；
            发生敲入后，派息方式不变，到期如果未敲出，结构为看跌空头或熊市价差空头
        FCN(Fixed Coupon Note, 固定派息票据)/DCN(Digital Coupon Note, 二元派息票据)/Phoenix凤凰票据
            FCN：无派息价格（派息价格为0），每月固定派息；到期日观察是否敲入
            DCN：有派息价格，每个观察日如果价格高于派息价格，派固定利息；到期日观察是否敲入
            Phoenix：有派息价格，每个观察日如果价格高于派息价格，派固定利息；每日观察是否敲入
    """

    def _init_boundary_condition(self, smax, maturity):
        """初始化边界条件
        Args:
            smax: float, 标的价格上界
            maturity: float, 到期时间，年化自然日期限
        Returns: None
        """
        prod = self.prod
        # 矩阵第一行: 标的价格为0时结构的价值
        t_vec = self.dt * self.j_vec
        interest_maturity_discount_factor = self.process.interest.disc_factor(maturity, t_vec)
        if np.any(prod.barrier_yield) > 0:
            # 如果派息价格不为0，则无派息, (预付金+敲入亏损)*贴现
            self.fd_not_in.v_grid[0, :] = self.fd_knockin.v_grid[0, :] = ((prod.strike_lower - prod.strike_upper)
                                                                          * prod.parti_in + prod.margin_lvl * prod.s0
                                                                          ) * interest_maturity_discount_factor
        else:
            # 如果派息价格为0，即FCN，则派息随着每个派息日累积，(累计派息+预付金+敲入亏损)*贴现
            self.fd_not_in.v_grid[0, self.t_step] = self.fd_knockin.v_grid[0, self.t_step] = (
                    (prod.strike_lower - prod.strike_upper)
                    * prod.parti_in + (prod.margin_lvl + self.next_coupon) * prod.s0)
            for j in range(self.t_step - 1, -1, -1):  # todo: 对常数无风险利率做优化，避免重复计算折现因子
                self.fd_not_in.v_grid[0, j] = self.fd_knockin.v_grid[0, j] = (self.fd_knockin.v_grid[0, j + 1]
                                                                              * self.process.interest.disc_factor(
                            j * self.dt, (j + 1) * self.dt))
                if j in self.out_dates:
                    self.fd_not_in.v_grid[0, j] = self.fd_knockin.v_grid[0, j] = (self.fd_knockin.v_grid[0, j]
                                                                                  + self.coupons[j] * prod.s0)

        # 矩阵最后一行: 每一个观察日前，股价上界能得到的本金加票息贴现
        self.next_paydate = self.pay_dates.repeat(
            np.diff(np.append(np.zeros((1,)), self.out_dates[::-1])).astype(int))  # 计息时间数
        self.next_paydate = np.append(self.pay_dates[0], self.next_paydate)  # 每个交易日对应的下一个敲出观察日对应的付息日
        # 每一个观察日前，股价上界能得到：(派息+预付金)*贴现。这里派息次数取经过的敲出观察日的次数，这是近似的取值，因为派息次数是路径依赖的。
        # 由于股价上界很大，近似认为之前的每次派息日，价格都足够高，每次都可以获得派息。
        interest_next_paydate_discount_factor = self.process.interest.disc_factor(self.next_paydate, t_vec)
        v_smax = (prod.margin_lvl + self.coupons) * prod.s0 * interest_next_paydate_discount_factor
        self.fd_knockin.v_grid[self.s_step, :] = self.fd_not_in.v_grid[self.s_step, :] = v_smax

    def _init_terminal_condition(self, s_vec, maturity):
        """初始化终止时已敲入且未敲出、未敲入且未敲出的期权价值
        Args:
            s_vec: List[float], 网格的价格格点
            maturity: float, 到期时间，年化自然日期限
        Returns: None
        """
        prod = self.prod
        # 初始化next_barrier_in、next_barrier_out、next_coupon为最后一期的敲入价、敲出价、敲出票息。若变敲出变敲入，需要动态调整
        # 设置到期日的敲入边界位置
        if isinstance(prod.barrier_in, (int, float, np.int32, np.float64)):
            # 常数敲入
            self.next_barrier_in = prod.barrier_in
        elif isinstance(prod.barrier_in, (list, np.ndarray)):
            # self.out_dates是估值日之后的、经过截断的敲出观察日，取产品参数列表中最后len(self.out_dates)个，与其对应
            self._barrier_in = prod.barrier_in[-len(self.out_dates):].copy()
            # 浮动敲入
            self.next_barrier_in = self._barrier_in[-1]
            self.reversed_barrier_in = np.flip(np.array(self._barrier_in))
        else:
            raise ValueError(f'敲入障碍价类型为{type(prod.barrier_in)}，仅支持int/float/list/np.ndarray，请检查')

        # 设置到期日的敲出边界位置
        if isinstance(prod.barrier_out, (int, float, np.int32, np.float64)):
            # 常数敲出
            self.next_barrier_out = prod.barrier_out
        elif isinstance(prod.barrier_out, (list, np.ndarray)):
            # self.out_dates是估值日之后的、经过截断的敲出观察日，取产品参数列表中最后len(self.out_dates)个，与其对应
            self._barrier_out = prod.barrier_out[-len(self.out_dates):].copy()
            # 阶梯式敲出
            self.next_barrier_out = self._barrier_out[-1]
            self.reversed_barrier_out = np.flip(np.array(self._barrier_out))
        else:
            raise ValueError(f'敲出障碍价类型为{type(prod.barrier_out)}，仅支持int/float/list/np.ndarray，请检查')

        # 设置到期日的派息边界位置
        if isinstance(prod.barrier_yield, (int, float, np.int32, np.float64)):
            # 常数派息
            self.next_barrier_yield = prod.barrier_yield
        elif isinstance(prod.barrier_yield, (list, np.ndarray)):
            # self.out_dates是估值日之后的、经过截断的敲出观察日，取产品参数列表中最后len(self.out_dates)个，与其对应
            self._barrier_yield = prod.barrier_yield[-len(self.out_dates):].copy()
            # 浮动派息
            self.next_barrier_yield = self._barrier_yield[-1]
            self.reversed_barrier_yield = np.flip(np.array(self._barrier_yield))

        # 敲出票息
        if isinstance(prod.coupon, (int, float, np.int32, np.float64)):
            # 常数敲出票息
            self.next_coupon = prod.coupon
            self.coupons = np.full(self.t_step + 1, prod.coupon)  # N天的敲出票息列表，用于生成mspot价格上边界条件
        elif isinstance(prod.coupon, (list, np.ndarray)):
            # self.out_dates是估值日之后的、经过截断的敲出观察日，取产品参数列表中最后len(self.out_dates)个，与其对应
            self._coupon = prod.coupon[-len(self.out_dates):].copy()
            # 浮动敲出票息
            self.next_coupon = self._coupon[-1]
            self.reversed_coupon = np.flip(np.array(self._coupon))
            # N天的敲出票息列表，用于生成smax价格上边界条件
            coupons = np.array(self._coupon).repeat(
                np.diff(np.append(np.zeros((1,)), self.out_dates[::-1])).astype(int))
            self.coupons = np.append(coupons[0], coupons)
        else:
            raise ValueError(f'敲出票息类型为{type(prod.coupon)}，仅支持int/float/list/np.ndarray，请检查')

        # in_idx：小于到期日敲入线的最大的标的价格索引。如果是变敲入，此变量in_idx应该随敲入价动态调整。
        self.in_idx = 1 + np.where(s_vec <= self.next_barrier_in)[0][-1] if self.next_barrier_in > 0 else 0
        # out_idxs：大于到期日敲出线的标的价格数组的索引。如果是变敲出，此变量out_idxs应该随敲出价动态调整。
        self.out_idxs = 1 + np.array(np.where(s_vec >= self.next_barrier_out)[0])
        # yld_idxs: 大于到期日派息边界位置的标的价格数组的索引。如果是变派息，此变量yld_idxs应该随敲出价动态调整。
        self.yld_idxs = 1 + np.array(np.where(s_vec >= self.next_barrier_yield)[0])

        # 矩阵最后一列：到期payoff
        # 已敲入：派息 + 敲入或有亏损 + 预付金
        self.fd_knockin.v_grid[1:-1, self.t_step] = np.where(s_vec >= self.next_barrier_yield,
                                                             self.next_coupon * prod.s0,
                                                             0) + \
                                                    np.where(s_vec > prod.strike_upper, 0,
                                                             (np.where(s_vec > prod.strike_lower, s_vec,
                                                                       prod.strike_lower) - prod.strike_upper) *
                                                             prod.parti_in) + prod.margin_lvl * prod.s0
        # 未敲入：派息 + 到期敲入或有亏损 + 预付金
        self.fd_not_in.v_grid[1:-1, self.t_step] = np.where(s_vec >= self.next_barrier_yield,
                                                            self.next_coupon * prod.s0,
                                                            0) + \
                                                   np.where(s_vec > self.next_barrier_in, 0,
                                                            # 敲入亏损
                                                            (np.where(s_vec > prod.strike_lower, s_vec,
                                                                      prod.strike_lower) - prod.strike_upper) *
                                                            prod.parti_in) + prod.margin_lvl * prod.s0

    def _backward_induction(self):
        """从倒数第二天向第一天逆向迭代"""
        prod = self.prod
        if prod.in_obs_type == ExerciseType.European:  # 敲入观察为欧式，仅到期观察敲入
            self.next_barrier_in = -1
            self.in_idx = 0
        for j in range(self.t_step - 1, -1, -1):
            self.fd_knockin.v_grid[1:-1, j] = self.fd_knockin.evolve(j, self.fd_knockin.v_grid[1:-1, j + 1], self.dt)
            self.fd_not_in.v_grid[1:-1, j] = self.fd_not_in.evolve(j, self.fd_not_in.v_grid[1:-1, j + 1], self.dt)

            # 考虑敲出的情况: 敲出观察日边界条件需更改
            if j in self.out_dates:

                # 如果敲出价是浮动的，调整敲出价
                self.next_barrier_out = (self.reversed_barrier_out[np.where(self.out_dates == j)[0][0]]
                                         if isinstance(prod.barrier_out, (list, np.ndarray)) else prod.barrier_out)
                self.out_idxs = 1 + np.array(np.where(self.fd_not_in.s_vec >= self.next_barrier_out)[0])
                # 如果派息边界是浮动的，调整派息边界
                self.next_barrier_yield = (self.reversed_barrier_yield[np.where(self.out_dates == j)[0][0]]
                                           if isinstance(prod.barrier_yield,
                                                         (list, np.ndarray)) else prod.barrier_yield)
                self.yld_idxs = 1 + np.array(np.where(self.fd_not_in.s_vec >= self.next_barrier_yield)[0])
                # 如果敲出票息是浮动的，调整敲出票息
                self.next_coupon = (self.reversed_coupon[np.where(self.out_dates == j)[0][0]]
                                    if isinstance(prod.coupon, (list, np.ndarray)) else prod.coupon)
                # 在票息边界和敲出边界之间，无敲入phoenix的价格为亏损贴现加上票息
                margin_s = prod.margin_lvl * prod.s0
                self.fd_knockin.v_grid[self.yld_idxs, j] = (self.next_coupon * prod.s0 + np.where(
                    self.fd_not_in.s_vec[self.yld_idxs - 1] < self.next_barrier_out,
                    self.fd_knockin.v_grid[self.yld_idxs, j], margin_s))
                self.fd_not_in.v_grid[self.yld_idxs, j] = (self.next_coupon * prod.s0 + np.where(
                    self.fd_not_in.s_vec[self.yld_idxs - 1] < self.next_barrier_out,
                    self.fd_not_in.v_grid[self.yld_idxs, j], margin_s))

                if prod.in_obs_type != ExerciseType.European:
                    # 如果敲入价是浮动的，调整敲入价
                    self.next_barrier_in = self.reversed_barrier_in[np.where(self.out_dates == j)[0][0]] if isinstance(
                        prod.barrier_in, (list, np.ndarray)) else prod.barrier_in
                    self.in_idx = 1 + np.where(self.fd_not_in.s_vec <= self.next_barrier_in)[0][
                        -1] if self.next_barrier_in > 0 else 0

            # 考虑敲入的情况: 向下越过敲入边界时，下一时间点敲入边界以上部分是标准autocall价格
            self.fd_not_in.v_grid[:(self.in_idx + 1), j] = self.fd_knockin.v_grid[:(self.in_idx + 1), j]



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.time import global_evaluation_date
from pricelib.common.utilities.enums import CallPut, ExerciseType, PaymentType
from pricelib.common.pricing_engine_base import McEngine


class MCDigitalEngine(McEngine):
    """二元(数字)期权-现金或无-Monte Carlo模拟定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date,
                                                                            prod.end_date)
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        r = self.process.interest(_maturity)
        paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()

        # 美式：
        if prod.exercise_type == ExerciseType.American:
            if prod.discrete_obs_interval is None:  # 每日观察
                obs_points = np.arange(0, _maturity_business_days + 1, 1)[1:]
            else:  # 均匀离散观察
                dt_step = prod.discrete_obs_interval * prod.t_step_per_year
                obs_points = np.flip(np.round(np.arange(_maturity_business_days, 0, -dt_step)).astype(int))
                obs_points = np.concatenate((np.array([0]), obs_points))
                obs_points[obs_points > _maturity_business_days] = _maturity_business_days  # 防止闰年导致的下标越界

            if prod.callput == CallPut.Call:
                strike_bool = np.where(paths[obs_points] >= prod.strike, 1, 0)
            else:  # self.callput == CallPut.Put:
                strike_bool = np.where(paths[obs_points] <= prod.strike, 1, 0)
            value = prod.rebate * np.max(strike_bool, axis=0)  # 能否拿到票息
            if prod.payment_type == PaymentType.Hit:
                strike_time = (np.argmax(strike_bool, axis=0) + 1) * (1 / prod.t_step_per_year)  # 停时
                value = np.mean(value * np.exp(-r * strike_time))
            else:  # self.payment_type == PaymentType.Expire:
                value = np.mean(value * math.exp(-r * _maturity))

        # 欧式：
        elif prod.exercise_type == ExerciseType.European:
            if prod.callput == CallPut.Call:
                payoff = (paths[-1] >= prod.strike) * prod.rebate
            else:  # prod.callput == CallPut.Put:
                payoff = (paths[-1] <= prod.strike) * prod.rebate
            value = np.mean(payoff) * math.exp(-r * _maturity)
        else:
            raise ValueError("观察方式只支持美式观察American或欧式观察European")

        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.time import global_evaluation_date


class MCPhoenixEngine(McEngine):
    """凤凰类 AutoCallable Monte Carlo模拟定价引擎，不支持变敲入、变敲出、变票息
            较雪球式AutoCallable，增加派息价格，少了敲出和红利票息
            每个派息（敲出）观察日，如果价格高于派息价格，派固定利息，如果发生敲出，合约提前结束；
            发生敲入后，派息方式不变，到期如果未敲出，结构为看跌空头
    """

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        obs_dates = prod.obs_dates.count_business_days(calculate_date)
        obs_dates = np.array([num for num in obs_dates if num >= 0])
        pay_dates = prod.pay_dates.count_calendar_days(calculate_date)
        pay_dates = np.array([num / prod.annual_days.value for num in pay_dates if num >= 0])
        # 经过估值日截断的列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
        _barrier_out = prod.barrier_out[-len(obs_dates):].copy()
        _barrier_in = prod.barrier_in[-len(obs_dates):].copy()
        _barrier_yield = prod.barrier_yield[-len(obs_dates):].copy()
        _coupon = prod.coupon[-len(obs_dates):].copy()
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()

        # 计算每条价格路径最小敲出时间
        barrier_out = np.tile(_barrier_out, (self.n_path, 1)).T
        # 每条路径敲出时间索引(第几个月敲出)，这里的矩阵是True、False
        barrier_out_bool = paths[obs_dates, :] > barrier_out
        # np.argmax返回的是每一列的第一个最大值True的索引
        knock_out_time_idx = np.argmax(barrier_out_bool, axis=0)
        # 如果一列中没有大于barrier的元素（全是False），那么argmax会返回0，因此需要将其‘索引’进行修正为paths.shape[0]，
        # 即最后一个观察日的索引再+1，例如两年24个月的最后一个观察日索引是23，则未敲出索引设为24
        knock_out_time_idx = np.where(np.any(barrier_out_bool, axis=0), knock_out_time_idx, obs_dates.size)
        # 统计哪些路径属于发生了敲出的情景(布尔索引)
        knock_out_scenario = np.where(knock_out_time_idx < obs_dates.size, True, False)
        # 每条路径持有时长，对于未敲出情形，持有到期，因此将索引设置为最后一个观察日的索引，例如23
        hold_time_idx = knock_out_time_idx.copy()
        hold_time_idx[~knock_out_scenario] = obs_dates.size - 1

        # 统计哪些派息日，标的价格在派息线下方(即该派息日不派息)
        barrier_yield = np.tile(_barrier_yield, (self.n_path, 1)).T
        no_coupon_time_idx = paths[obs_dates, :] < barrier_yield
        # 将发生敲出之后的未派息bool由True改为False
        no_coupon_bool = np.where(np.arange(no_coupon_time_idx.shape[0])[:, np.newaxis] > knock_out_time_idx, False,
                                  no_coupon_time_idx)

        # 排除发生了敲出的路径，统计哪些路径属于敲入未敲出
        barrier_in = np.tile(_barrier_in, (self.n_path, 1)).T
        knock_in_time_idx = paths[obs_dates, :] < barrier_in
        # 将发生敲出之后的敲入由True改为False
        knock_in_bool = np.where(np.arange(knock_in_time_idx.shape[0])[:, np.newaxis] > knock_out_time_idx, False,
                                 knock_in_time_idx)
        # 统计哪些路径属于敲入未敲出
        knock_in_scenario = np.argmax(knock_in_bool, axis=0) > 0  # 敲入未敲出情形

        # 不同派息日的票息分别计算贴现因子，得到派息的现值
        discount_factor = np.empty(pay_dates.size)
        for i, pay_d in enumerate(pay_dates):
            discount_factor[i] = self.process.interest.disc_factor(pay_d)
        discounted_coupon = _coupon * prod.s0 * discount_factor
        # 将每个月的派息现值累加到一起
        cumulated_coupon = np.cumsum(discounted_coupon)

        # payoff汇总
        payoff = 0
        # 1.未敲入的部分（敲出/未敲入未敲出），到期还本
        payoff += np.sum(discount_factor[hold_time_idx[~knock_in_scenario]]) * prod.margin_lvl * prod.s0
        # 2.派息payoff
        # 2.1 首先忽略派息线以下，假设在敲出之前的每个派息日都能拿到派息，将派息现值累加
        payoff += np.sum(cumulated_coupon[hold_time_idx])
        # 2.2 然后减去处于派息线下方的派息
        payoff -= np.sum(np.sum(no_coupon_bool, axis=1) * discounted_coupon)
        # 3.敲入，承担跌幅损失
        s_vec = paths[-1, knock_in_scenario].copy()
        s_vec[np.where(s_vec < prod.strike_lower)] = prod.strike_lower
        payoff += ((prod.margin_lvl * prod.s0 - prod.strike_upper) * np.sum(knock_in_scenario
                                                                            ) + np.sum(s_vec)) * discount_factor[-1]

        payoff /= self.n_path
        return payoff



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from ..utilities.enums import VolType
from ..utilities.patterns import Observable
from ..utilities.utility import logging


class BlackConstVol(Observable):
    """BSM常数波动率，被观察者"""
    vol_type = VolType.CV

    def __init__(self, vol, name=None):
        super().__init__()
        self.__volval = vol
        self.__name = name

    def __call__(self, *args, **kwargs):
        return self.__volval

    @property
    def volval(self):
        return self.__volval

    @volval.setter
    def volval(self, new_value):
        """覆写给对象的属性赋值的方法。在被观察者的值改变时，自动向观察者发送通知"""
        try:
            self.__volval = float(new_value)  # 被观察者值变化
        except ValueError as e:
            logging.error(f"ValueError:{e}")
        else:
            self.notify_observers(new_value, name=self.__name)  # 自动向观察者发送通知



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .autocall import AutoCall
from .autocallable_base import AutocallableBase, PhoenixBase
from .standard_snowball import StandardSnowball
from .earlyprofit_snowball import EarlyProfitSnowball
from .butterfly_snowball import ButterflySnowball
from .stepdown_snowball import StepDownSnowball
from .parachute_snowball import ParachuteSnowball
from .otm_snowball import OTMSnowball
from .bothdown_snowball import BothDownSnowball
from .snowball_plus import SnowballPlus
from .floored_snowball import FlooredSnowball
from .phoenix import Phoenix
from .fcn import FCN
from .dcn import DCN
from .paris_snowball import ParisSnowball



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.utilities.enums import InOut, ExerciseType, PaymentType
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import McEngine


class MCDoubleBarrierEngine(McEngine):
    """双边障碍期权 Monte Carlo 模拟定价引擎
    支持欧式观察(仅到期观察)/美式观察(整个有效期观察)；只支持离散观察(默认为每日观察)；
    支持现金返还；敲入现金返还为到期支付；敲出现金返还支持 立即支付/到期支付"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        r = self.process.interest(_maturity)
        paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()
        payoff = prod.callput.value * (paths[-1] - prod.strike) * prod.parti
        # 美式：
        if prod.exercise_type == ExerciseType.American:
            if prod.discrete_obs_interval is None:  # 每日观察
                obs_points = np.arange(0, _maturity_business_days + 1, 1)[1:]
            else:  # 均匀离散观察
                dt_step = prod.discrete_obs_interval * prod.t_step_per_year
                obs_points = np.flip(np.round(np.arange(_maturity_business_days, 0, -dt_step)).astype(int))
                obs_points[obs_points > _maturity_business_days] = _maturity_business_days  # 防止闰年导致的下标越界
            if prod.inout == InOut.In:  # 美式双边敲入一定是到期支付
                knock_in_bool = np.any((paths[obs_points] <= prod.bound[0]) | (paths[obs_points] >= prod.bound[1]),
                                       axis=0)
                value = np.where(knock_in_bool, np.where(payoff < 0, 0, payoff), prod.rebate[0])
                value = np.mean(value) * math.exp(-r * _maturity)
            elif prod.inout == InOut.Out:
                step_index = np.tile(obs_points, (self.n_path, 1)).T
                hit_lower = np.min(
                    np.where(paths[1:] <= prod.bound[0], step_index.astype(int), _maturity_business_days + 1),
                    axis=0)
                hit_upper = np.min(
                    np.where(paths[1:] >= prod.bound[1], step_index.astype(int), _maturity_business_days + 1),
                    axis=0)
                if prod.payment_type == PaymentType.Hit:
                    value = np.mean(
                        np.where(hit_lower < hit_upper, prod.rebate[0] * np.exp(-r * hit_lower / prod.t_step_per_year),
                                 np.where(hit_lower > hit_upper,
                                          prod.rebate[1] * np.exp(-r * hit_upper / prod.t_step_per_year),
                                          np.where(payoff < 0, 0, payoff) * math.exp(-r * _maturity))))
                elif prod.payment_type == PaymentType.Expire:
                    value = np.mean(np.where(hit_lower < hit_upper, prod.rebate[0],
                                             np.where(hit_lower > hit_upper, prod.rebate[1],
                                                      np.where(payoff < 0, 0, payoff)))) * math.exp(-r * _maturity)
                else:
                    raise ValueError("PaymentType must be Hit or Expire")
            else:
                raise ValueError("InOut must be In or Out")
        # 欧式：
        elif prod.exercise_type == ExerciseType.European:
            if prod.inout == InOut.Out:
                value = np.where(paths[-1] <= prod.bound[0], prod.rebate[0],
                                 np.where(paths[-1] >= prod.bound[1], prod.rebate[1],
                                          np.where(payoff < 0, 0, payoff)))
            elif prod.inout == InOut.In:
                value = np.where((paths[-1] > prod.bound[0]) & (paths[-1] < prod.bound[1]), prod.rebate[0],
                                 np.where(payoff < 0, 0, payoff))
            else:
                raise ValueError("InOut must be In or Out")
            value = np.mean(value) * math.exp(-r * _maturity)
        else:
            raise ValueError("ExerciseType must be American or European")
        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .vanilla import *
from .digital import *
from .barrier import *
from .asian import *
from .autocallable import *
from .cashflow import *
from .accurals import *



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .analytic_engine_base import AnalyticEngine
from .mc_engine_base import McEngine
from .pde_engine_base import FdmEngine, FdmGrid, FdmGridwithBound
from .quad_engine_base import QuadEngine
from .tree_engine_base import BiTreeEngine

__all__ = ['AnalyticEngine', 'McEngine', 'FdmEngine', 'FdmGrid', 'FdmGridwithBound', 'QuadEngine', 'BiTreeEngine']



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from enum import Enum, unique  # @unique装饰器可以防止枚举成员具有相同的值
from collections import namedtuple  # 具名元组，实现枚举成员的自定义属性


@unique
class CallPut(Enum):  # 看涨/看跌
    Call = 1
    Put = -1


@unique
class BuySell(Enum):  # 买入/卖出
    Buy = 1
    Sell = -1


@unique
class InOut(Enum):  # 敲入/敲出
    In = "敲入"
    Out = "敲出"


@unique
class TouchType(Enum):  # 触碰/不触碰
    Touch = "触碰"
    NoTouch = "不触碰"


@unique
class UpDown(Enum):  # 向上/向下
    Up = 1
    Down = -1


# 具名元组: 障碍期权类型
BarrierTuple = namedtuple(typename='BarrierTuple', field_names=['text', 'updown', 'inout', 'callput'])


@unique
class BarrierType(Enum):  # 障碍期权类型
    UOC = BarrierTuple("向上敲出看涨", UpDown.Up, InOut.Out, CallPut.Call)
    UOP = BarrierTuple("向上敲出看跌", UpDown.Up, InOut.Out, CallPut.Put)
    DOC = BarrierTuple("向下敲出看涨", UpDown.Down, InOut.Out, CallPut.Call)
    DOP = BarrierTuple("向下敲出看跌", UpDown.Down, InOut.Out, CallPut.Put)
    UIC = BarrierTuple("向上敲入看涨", UpDown.Up, InOut.In, CallPut.Call)
    UIP = BarrierTuple("向上敲入看跌", UpDown.Up, InOut.In, CallPut.Put)
    DIC = BarrierTuple("向下敲入看涨", UpDown.Down, InOut.In, CallPut.Call)
    DIP = BarrierTuple("向下敲入看跌", UpDown.Down, InOut.In, CallPut.Put)

    @property
    def updown(self):
        return self.value.updown

    @property
    def inout(self):
        return self.value.inout

    @property
    def callput(self):
        return self.value.callput

    @staticmethod
    def get_category(updown, input, callput):
        return BARRIER_TYPE_MAPPING[f"{updown.value}{input.value}{callput.value}"]


BARRIER_TYPE_MAPPING = {"1敲出1": BarrierType.UOC, "1敲出-1": BarrierType.UOP, "-1敲出1": BarrierType.DOC,
                        "-1敲出-1": BarrierType.DOP, "1敲入1": BarrierType.UIC, "1敲入-1": BarrierType.UIP,
                        "-1敲入1": BarrierType.DIC, "-1敲入-1": BarrierType.DIP}


@unique
class RandsMethod(Enum):  # 随机数生成方法
    Pseudorandom = "伪随机"
    LowDiscrepancy = "低差异"


@unique
class LdMethod(Enum):  # 低差异序列生成方法
    Halton = "Halton"
    Sobol = "Sobol"


@unique
class VolType(Enum):  # 波动率类型
    CV = "常数波动率"
    LV = "局部波动率"
    SV = "随机波动率"
    SLV = "随机局部波动率"


@unique
class ProcessType(Enum):  # 目前支持的随机过程类型
    BSProcess1D = "一维BS随机过程"
    Heston = "Heston随机过程"


@unique
class EngineType(Enum):  # 定价引擎类型
    AnEngine = "解析法定价引擎"
    McEngine = "蒙特卡洛模拟定价引擎"
    TreeEngine = "树方法定价引擎"
    PdeEngine = "PDE方法定价引擎"
    QuadEngine = "数值积分定价引擎"


@unique
class QuadMethod(Enum):  # 数值积分方法
    Trapezoid = "梯形法则"
    Simpson = "Simpson法则"


@unique
class ExerciseType(Enum):  # 行权方式
    European = "欧式"
    American = "美式"
    Asian = "亚式"
    Bermudan = "百慕大"


@unique
class PaymentType(Enum):  # 支付方式
    Expire = "到期支付"
    Hit = "立即支付"


@unique
class AverageMethod(Enum):
    Arithmetic = "算术平均"
    Geometric = "几何平均"


@unique
class AsianAveSubstitution(Enum):
    Underlying = "代替标的结算价"
    Strike = "代替执行价"


@unique
class StatusType(Enum):
    DownTouch = "敲入"
    UpTouch = "敲出"
    NoTouch = "未敲入未敲出"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import ABCMeta
import numpy as np
from scipy.stats import norm
from ..processes import StochProcessBase
from ..utilities.enums import ProcessType, EngineType
from .engine_base import PricingEngineBase


class BiTreeEngine(PricingEngineBase, metaclass=ABCMeta):
    """二叉树定价引擎"""
    engine_type = EngineType.TreeEngine

    def __init__(self, stoch_process: StochProcessBase = None, tree_branches=500, *,
                 s=None, r=None, q=None, vol=None):
        """初始化二叉树定价引擎
        Args:
            stoch_process: StochProcessBase，随机过程
            tree_branches: int，二叉树每年的分支数
       在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.tree_branches = tree_branches  # 二叉树每年的分支数
        self.dt = 1 / tree_branches  # 时间步长
        # 以下为计算过程的中间变量
        self.p = None  # 风险中性概率，在path_generator方法中计算生成
        self.s_paths = None  # CRR二叉树
        self.v = None  # 期权价值矩阵
        self.r = None  # 无风险利率
        # 以下变量仅当需要使用Vma、Vmb这两个解析解方法的时候才需要在子类中赋值
        self.q = None  # 贴水率
        self.vol = None  # 波动率

    def set_stoch_process(self, stoch_process: StochProcessBase):
        """设置随机过程"""
        assert stoch_process.process_type == ProcessType.BSProcess1D, 'Error: 二叉树方法只能使用1维BSM动态过程'
        self.process = stoch_process

    def path_generator(self, n_step, t=None, spot=None):
        """生成上三角结构的标的资产价格CRR二叉树
        Args:
            n_step: int，二叉树的分支数
            t: float，到期时间
            spot: float，期初价格
        Returns:
            self.s_paths: np.array，上三角结构的标的资产价格CRR二叉树
        """
        if spot is None:
            spot = self.process.spot()
        u = np.exp(self.process.diffusion(t, spot) * np.sqrt(self.dt))
        d = 1 / u
        self.p = (np.exp(self.process.drift(t) * self.dt) - d) / (u - d)  # 风险中性概率
        u_power = np.arange(n_step + 1)
        u_power_matrix = np.tile(u_power, (n_step + 1, 1))
        d_power_matrix = u_power_matrix.T
        u_matrix = u ** u_power_matrix
        d_matrix = d ** d_power_matrix
        self.s_paths = u_matrix * d_matrix
        for i in range(n_step + 1):
            self.s_paths[i] = np.roll(self.s_paths[i], i)
        self.s_paths *= spot
        self.s_paths = np.triu(self.s_paths, k=0)
        return self.s_paths

    def Vma(self, s, km, epsilon, t):
        """ Vma = 欧式香草期权中的资产或无的价值
        例如看涨期权，到期时刻，St > strike, Vma = St; St < strike, Vma = 0
        Args:
            s: 期初价格
            km: 行权价
            epsilon: 1 or -1，看涨期权为1，看跌期权为-1
            t: 年化到期时间
        Returns: Vma = 欧式香草期权中的资产或无的部分
        """
        tau = 0.5 * self.vol ** 2 * t
        d1 = (np.log(s / km) + (self.r - self.q + 0.5 * self.vol ** 2) * t) / (self.vol * np.sqrt(t))
        return np.exp((-2 * self.q * tau) / (self.vol ** 2)) * s * norm.cdf(epsilon * d1)

    def Vmb(self, s, km, epsilon, t):
        """ Vmb = 欧式香草期权中的现金或无的价值
        例如看涨期权，到期时刻，St > strike, Vmb = 1 ; St < strike, Vmb = 0
        Args:
            s: 期初价格
            km: 行权价
            epsilon: 1 or -1，看涨期权为1，看跌期权为-1
            t: 年化到期时间
        Returns: Vmb = 欧式香草期权中的现金或无的部分
        """
        tau = 0.5 * self.vol ** 2 * t
        d1 = (np.log(s / km) + (self.r - self.q + 0.5 * self.vol ** 2) * t) / (self.vol * np.sqrt(t))
        d2 = d1 - self.vol * np.sqrt(t)
        return np.exp((-2 * self.r * tau) / (self.vol ** 2)) * norm.cdf(epsilon * d2)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.pricing_engine_base import QuadEngine
from pricelib.common.time import global_evaluation_date


class QuadPortfolioEngine(QuadEngine):
    """香草组合积分法定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        tau = (prod.end_date - calculate_date).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)
        self._check_method_params()

        # 设定期末价值
        self.init_grid(spot, vol, tau)
        v_vec = np.zeros(self.ln_s_vec.shape)
        for vanilla, position in prod.vanilla_list:
            v_vec += np.maximum(vanilla.callput.value * (np.exp(self.ln_s_vec) - vanilla.strike), 0) * position
        # 设置积分法engine参数
        self.set_quad_params(r=r, q=q, vol=vol)
        # 回溯计算
        return self.fft_step_backward(np.log(np.array([spot])), self.ln_s_vec, v_vec, tau)[0]



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import CallPut, InOut, UpDown, PaymentType, ExerciseType
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import AnalyticEngine
from ...products.digital.digital_option import DigitalOption
from ...products.barrier.barrier_option import BarrierOption
from .analytic_barrier_engine import AnalyticBarrierEngine
from .analytic_asset_or_nothing_engine import AnalyticAssetOrNothingEngine
from .analytic_cash_or_nothing_engine import AnalyticCashOrNothingEngine


class AnalyticAirbagEngine(AnalyticEngine):
    """安全气囊闭式解定价引擎
    安全气囊是美式观察向下敲入看涨期权，敲入前是call，敲入后payoff与持有标的资产相同(在参与率为1的情况下)
    相当于向下敲出看涨期权 + 向下触碰资产或无期权 的组合
    因为airbag解析解由障碍与二元解析解组合而成，所以敲入的下跌参与率与重置后看涨参与率必须相等"""

    # pylint: disable=invalid-name
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        assert prod.knockin_parti == prod.reset_call_parti, "Error: 安全气囊解析解由障碍与二元解析解组合而成，敲入的下跌参与率与重置后看涨参与率必须相等"
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        vol = self.process.vol(tau, self.process.spot())

        if prod.discrete_obs_interval is not None:
            # 均匀离散观察障碍期权，M. Broadie, P. Glasserman, S.G. Kou(1997) 在连续障碍期权解析解上加调整项，调整障碍价格水平
            # 指数上的beta = -zeta(1/2) / sqrt(2*pi) = 0.5826, 其中zeta是黎曼zeta函数
            if prod.updown == UpDown.Up:  # 向上
                barrier = prod.barrier * np.exp(0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            elif prod.updown == UpDown.Down:  # 向下
                barrier = prod.barrier * np.exp(-0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            else:
                raise ValueError(f"Error: {prod} 无效的updown参数: {prod.updown}")
        else:  # 连续观察障碍期权
            barrier = prod.barrier

        DOC = BarrierOption(strike=prod.strike, barrier=barrier, rebate=0, inout=InOut.Out,
                            updown=UpDown.Down, callput=CallPut.Call, start_date=prod.start_date,
                            end_date=prod.end_date, discrete_obs_interval=None,
                            engine=AnalyticBarrierEngine(self.process))
        DIAsset = DigitalOption(strike=barrier, rebate=0, callput=CallPut.Put, start_date=prod.start_date,
                                end_date=prod.end_date, discrete_obs_interval=None, exercise_type=ExerciseType.American,
                                payment_type=PaymentType.Expire, engine=AnalyticAssetOrNothingEngine(self.process))
        DICash = DigitalOption(strike=barrier, rebate=-prod.strike, callput=CallPut.Put, start_date=prod.start_date,
                               end_date=prod.end_date, discrete_obs_interval=None, exercise_type=ExerciseType.American,
                               payment_type=PaymentType.Expire, engine=AnalyticCashOrNothingEngine(self.process))
        DOC_value = DOC.price(t=t, spot=spot) * prod.call_parti
        # 期权无本金，票据有本金。合约设立时为平值，这里加上rebate为负的行权价的二元看跌期权，相当于减去本金
        DIAsset_value = (DIAsset.price(t=t, spot=spot) + DICash.price(t=t, spot=spot)) * prod.knockin_parti
        return DOC_value + DIAsset_value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from scipy.stats import norm

from pricelib.common.utilities.enums import CallPut, InOut, UpDown
from pricelib.common.processes import StochProcessBase
from pricelib.common.pricing_engine_base import AnalyticEngine
from pricelib.common.time import global_evaluation_date
from ...products.vanilla.vanilla_option import VanillaOption
from ...products.barrier.barrier_option import BarrierOption
from .analytic_vanilla_european_engine import AnalyticVanillaEuEngine
from .analytic_barrier_engine import AnalyticBarrierEngine


class AnalyticDoubleBarrierEngine(AnalyticEngine):
    """双边障碍期权闭式解定价引擎
    双边敲出期权，到期时间 T 之前，满足L<S<U，则到期回报为max(S-K,0)，否则到期回报为0。双边敲入期权等于香草期权多头与双边敲出期权空头的组合。
    注意: 没有现金返还，即rebate=0
    Haug(1998) Put-Call Barrier Transformations 对于标的持有成本为0的期权(期货期权)，利用认购-认沽障碍期权对称性，
                                                利用单边障碍期权解析解的组合给双障碍期权定价。
    Ikeda and Kunitomo(1992)将双障碍期权表示为加权正态分布函数的无限集，但是仅当行权价在障碍范围内时，公式才是成立的。
    Haug(1998)的解析解与Ikeda and Kunitomo(1992)的结果几乎一致，两者对比，
        Ikeda and Kunitomo(1992)适用于标的持有成本不为0的情形, Haug(1998)适用于行权价在障碍范围之外的情形。
    """

    def __init__(self, stoch_process: StochProcessBase = None, formula_type="Ikeda&Kunitomo1992",
                 series_num=10, delta1=0, delta2=0, *,
                 s=None, r=None, q=None, vol=None):
        """
        Args:
            stoch_process: 随机过程StochProcessBase对象
            self.formula_type: 解析解类型，Ikeda&Kunitomo1992或Haug1998
        其他三个参数是Ikeda&Kunitomo1992的参数：
            self.series_num: 用于计算近似解的级数项 默认 i = -10 ~ 10
            self.delta1: 上边界的曲率
            self.delta2: 下边界的曲率，
                当delta1 = delta2 = 0时，对应两条平行边界
                当delta1 < 0 < delta2 时，对应下边界随时间呈指数增长，而上百年姐随时间呈指数衰减
                当delta1 > 0 > delta2 时，对应一个向下凸的下边界和一个向上凸的上边界
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.formula_type = formula_type
        self.series_num = series_num
        self.i_vec = np.array(range(- series_num, series_num + 1)).astype(float)  # 用于计算近似解的级数项
        self.delta1 = delta1
        self.delta2 = delta2
        # 以下是计算过程的中间变量
        self.prod = None  # Product产品对象

    # pylint: disable=invalid-name, too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)

        vanilla_option = VanillaOption(maturity=tau, strike=prod.strike, callput=prod.callput,
                                       engine=AnalyticVanillaEuEngine(self.process))

        if prod.discrete_obs_interval is not None:
            # 均匀离散观察，M. Broadie, P. Glasserman, S.G. Kou(1997) 在连续障碍期权解析解上加调整项，调整障碍价格水平
            # 指数上的beta = -zeta(1/2) / sqrt(2*pi) = 0.5826, 其中zeta是黎曼zeta函数
            bound = [None, None]
            bound[1] = prod.bound[1] * np.exp(0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            bound[0] = prod.bound[0] * np.exp(-0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
        else:  # 连续观察
            bound = prod.bound

        if self.formula_type == "Ikeda&Kunitomo1992":
            assert (bound[0] < prod.strike < bound[1]), "Error: Ikeda and Kunitomo(1992)的解析解仅当行权价在障碍范围内时，公式才是成立的。"
            return self.Ikeda_Kunitomo1992_pv(spot, r, q, vol, tau, bound, vanilla_option)
        if self.formula_type == "Haug1998":
            assert (r - q == 0), "Error: Haug(1998)的解析解仅当标的持有成本为0时(期货期权，或r-q=0)，公式才是准确的。"
            return self.Haug1998_pv(tau, bound, vanilla_option)
        raise NotImplementedError(f"{self.formula_type}不支持的解析解类型, 仅支持Ikeda&Kunitomo1992和Haug1998")

    # pylint: disable=invalid-name, too-many-locals
    def Ikeda_Kunitomo1992_pv(self, spot, r, q, vol, tau, bound, vanilla_option):
        """Ikeda and Kunitomo(1992)级数解，仅当行权价在障碍范围内时，公式才是成立的"""
        if self.prod.callput == CallPut.Call:
            E = bound[1] * math.exp(self.delta1 * tau)
        else:
            E = bound[0] * math.exp(self.delta2 * tau)
        numerator = (r - q + 0.5 * vol ** 2) * tau
        denominator = vol * math.sqrt(tau)
        d1 = (np.log(
            spot * np.power(bound[1], 2 * self.i_vec) / (self.prod.strike * np.power(bound[0], 2 * self.i_vec)))
              + numerator) / denominator
        d2 = (np.log(spot * np.power(bound[1], 2 * self.i_vec) / (E * np.power(bound[0], 2 * self.i_vec)))
              + numerator) / denominator
        d3 = (np.log(
            np.power(bound[0], 2 * self.i_vec + 2) / (self.prod.strike * spot * np.power(bound[1], 2 * self.i_vec)))
              + numerator) / denominator
        d4 = (np.log(
            np.power(bound[0], 2 * self.i_vec + 2) / (E * spot * np.power(bound[1], 2 * self.i_vec)))
              + numerator) / denominator
        mu1 = 2 * (r - q - self.delta2 - self.i_vec * (self.delta1 - self.delta2)) / vol ** 2 + 1
        mu2 = 2 * self.i_vec * (self.delta1 - self.delta2) / vol ** 2
        mu3 = 2 * (r - q - self.delta2 + self.i_vec * (self.delta1 - self.delta2)) / vol ** 2 + 1

        Nd1a = (np.power(np.power(bound[1], self.i_vec) / np.power(bound[0], self.i_vec), mu1) * (
                bound[0] / spot) ** mu2 * (norm.cdf(d1) - norm.cdf(d2)) * self.prod.callput.value)
        Nd1b = (np.power(np.power(bound[0], self.i_vec + 1) / (np.power(bound[1], self.i_vec) * spot), mu3) * (
                norm.cdf(d3) - norm.cdf(d4)) * self.prod.callput.value)
        Nd2a = (np.power(np.power(bound[1], self.i_vec) / np.power(bound[0], self.i_vec), mu1 - 2) * (
                bound[0] / spot) ** mu2 * self.prod.callput.value * (
                        norm.cdf(d1 - denominator) - norm.cdf(d2 - denominator)))
        Nd2b = (np.power(np.power(bound[0], self.i_vec + 1) / (np.power(bound[1], self.i_vec) * spot), mu3 - 2) * (
                norm.cdf(d3 - denominator) - norm.cdf(d4 - denominator)) * self.prod.callput.value)

        double_out = (self.prod.callput.value * spot * math.exp(-q * tau) * np.sum(Nd1a - Nd1b) -
                      self.prod.callput.value * self.prod.strike * math.exp(-r * tau) * np.sum(Nd2a - Nd2b))
        if self.prod.inout == InOut.Out:
            return double_out * self.prod.parti
        # else: self.inout == InOut.In
        double_in = vanilla_option.price() - double_out
        return double_in * self.prod.parti

    # pylint: disable=too-many-locals
    def Haug1998_pv(self, tau, bound, vanilla_option):
        """Haug(1998)的解析解，仅当标的持有成本为0时(期货期权，或r-q=0)，公式才是准确的"""
        value = 0
        if self.prod.callput == CallPut.Call:
            opposite_callput = CallPut.Put
        else:
            opposite_callput = CallPut.Call

        up_barrier = BarrierOption(maturity=tau, strike=self.prod.strike, barrier=bound[1],
                                   updown=UpDown.Up, callput=self.prod.callput, inout=InOut.In, rebate=0,
                                   discrete_obs_interval=None,
                                   engine=AnalyticBarrierEngine(self.process))
        down_barrier = BarrierOption(maturity=tau, strike=self.prod.strike, barrier=bound[0],
                                     updown=UpDown.Down, callput=self.prod.callput, inout=InOut.In, rebate=0,
                                     discrete_obs_interval=None,
                                     engine=AnalyticBarrierEngine(self.process))
        up_barrier_opposite = BarrierOption(strike=bound[1] ** 2 / self.prod.strike, barrier=bound[1] ** 2 / bound[0],
                                            maturity=tau, rebate=0, inout=InOut.In,
                                            updown=UpDown.Up, callput=opposite_callput,
                                            discrete_obs_interval=None,
                                            engine=AnalyticBarrierEngine(self.process))
        down_barrier_opposite = BarrierOption(strike=bound[0] ** 2 / self.prod.strike, barrier=bound[0] ** 2 / bound[1],
                                              maturity=tau, rebate=0, inout=InOut.In,
                                              updown=UpDown.Down, callput=opposite_callput,
                                              discrete_obs_interval=None,
                                              engine=AnalyticBarrierEngine(self.process))
        value += (up_barrier.price() + down_barrier.price()

                  - up_barrier_opposite.price() * self.prod.strike / bound[1]
                  - down_barrier_opposite.price() * self.prod.strike / bound[0])

        up_barrier.strike = self.prod.strike * bound[1] ** 2 / bound[0] ** 2
        up_barrier.barrier = math.pow(bound[1], 3) / bound[0] ** 2
        down_barrier.strike = self.prod.strike * bound[0] ** 2 / bound[1] ** 2
        down_barrier.barrier = math.pow(bound[0], 3) / bound[1] ** 2
        up_barrier_opposite.strike = math.pow(bound[1], 4) / (bound[0] ** 2 * self.prod.strike)
        up_barrier_opposite.barrier = math.pow(bound[1], 4) / math.pow(bound[0], 3)
        down_barrier_opposite.strike = math.pow(bound[0], 4) / (bound[1] ** 2 * self.prod.strike)
        down_barrier_opposite.barrier = math.pow(bound[0], 4) / math.pow(bound[1], 3)

        value += (up_barrier.price() * bound[0] / bound[1] + down_barrier.price() * bound[1] / bound[0]
                  - up_barrier_opposite.price() * self.prod.strike * bound[0] / bound[1] ** 2
                  - down_barrier_opposite.price() * self.prod.strike * bound[1] / bound[0] ** 2)

        up_barrier.strike = self.prod.strike * (math.pow(bound[1], 4) / math.pow(bound[0], 4))
        up_barrier.barrier = math.pow(bound[1], 5) / math.pow(bound[0], 4)
        down_barrier.strike = self.prod.strike * (math.pow(bound[0], 4) / math.pow(bound[1], 4))
        down_barrier.barrier = math.pow(bound[0], 5) / math.pow(bound[1], 4)

        value += (up_barrier.price() * bound[0] ** 2 / bound[1] ** 2
                  + down_barrier.price() * bound[1] ** 2 / bound[0] ** 2)

        double_in = min(value, vanilla_option.price())
        if self.prod.inout == InOut.In:
            return double_in * self.prod.parti
        # else: self.inout == InOut.Out
        double_out = vanilla_option.price() - double_in
        return double_out * self.prod.parti



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import time
from functools import wraps
import logging
import numpy as np

logging.basicConfig(format='%(asctime)s - %(levelname)s: %(message)s',  # - %(pathname)s[line:%(lineno)d]
                    level=logging.INFO)


def time_this(fn):
    """
    报告函数执行耗时的装饰器.
    @wraps(fn)使得@time_this不改变使用装饰器原有函数的结构(如__name__, __doc__)，避免fn.__name__返回"wrapper"，而不是fn本身的名字。
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        logging.debug('开始运行%s', fn.__name__)
        start = time.time()
        result = fn(*args, **kwargs)
        end = time.time()
        logging.info("%s执行完成，耗时%.3f秒", fn.__name__, end - start)
        return result

    return wrapper


def ascending_pairs(*args):
    """将多个时间入参*args拼接成数组并去重、排序，例如[t0, t1, t2, t3, t4] 然后打包成升序对 [(t0,t1),(t1,t2),(t2,t3),(t3,t4)]"""
    steps = np.sort(np.unique(np.hstack(args)))  # np.hstack 用于将数组按水平方向进行拼接，然后去重，排序
    return list(zip(steps[:-1], steps[1:]))


def descending_pairs(*args):
    """将多个时间入参*args接成数组并去重、排序，例如[t0, t1, t2, t3, t4] 然后打包成降序对 [(t4,t3),(t3,t2),(t2,t1),(t1,t0)]"""
    steps = np.sort(np.unique(np.hstack(args)))[::-1]
    return list(zip(steps[:-1], steps[1:]))


def descending_pairs_for_barrier(*args):
    """将多个时间入参*args接成数组并去重、排序，例如[t0, t1, t2, t3, t4] 然后打包成降序对 [(t4,t3),(t3,t2),(t2,t1),(t1,t0),(t0,t0)
    最后添加一个多余的(t0,t0)项, 因为障碍期权定价，最后定期初价格时，需要从df_bound切换到fd_full"""
    steps = np.sort(np.unique(np.hstack(args)))[::-1]
    steps = np.hstack((steps, steps[-1]))
    return list(zip(steps[:-1], steps[1:]))


def const_class(cls):
    """装饰器，修改常量类的属性，使其中的属性必须大写，并且值不能被更改"""

    @wraps(cls)
    def new_setattr(self, name, value):
        """重写设置属性方法
        允许加入新的常量，但不能更改已存在的常量的值
        新增常量的所有的字母需要大写
        Args:
            self: 常量类-对象
            name: 常量名，所有字母需要大写
            value: 常量值
        Returns: None
        """
        # # 不允许增加或修改任何常量的值
        # raise Exception('const : {} can not be changed'.format(name))
        # 允许加入新的常量，但不能更改已存在常量的值
        if name in self.__dict__:  # 若该常量已存在，不能二次赋值
            raise Exception(f"Can't change const {name}")
        if not name.isupper():  # 新增常量的所有的字母需要大写
            raise Exception(f"const name {name} is not all uppercase")
        self.__dict__[name] = value  # 允许加入新的常量，但不能更改值

    cls.__setattr__ = new_setattr
    return cls



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from ..utilities.enums import VolType


class StochLocalVol:
    """随机局部波动率模型基类"""
    vol_type = VolType.SLV


class HestonSLV(StochLocalVol):
    """Heston随机局部波动率模型"""

    def __init__(self):
        pass



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType
from .mc_autocallable_engine import MCAutoCallableEngine


class MCParisSnowballEngine(MCAutoCallableEngine):
    """巴黎雪球MC定价引擎"""

    def _cal_knock_in_scenario(self, calculate_date):
        """计算每条路径敲入时间"""
        prod = self.prod
        if prod.status == StatusType.DownTouch:
            pass
        else:
            knock_in_obs_dates = prod.knock_in_obs_dates.count_business_days(calculate_date)
            knock_in_obs_dates = np.round(np.array([num for num in knock_in_obs_dates if num >= 0])).astype(int)
            # 判断某一条路径是否有敲入
            knock_in_level = np.tile(self._barrier_in, (self.n_path, 1)).T  # 变敲入（与敲出观察日同长度的列表）
            knock_in_bool = np.where(self.s_paths[knock_in_obs_dates - 1] < knock_in_level, 1, 0)
            knock_in_count = np.cumsum(knock_in_bool, axis=0)
            knock_in_scenario = np.max(knock_in_count, axis=0) >= prod.knock_in_times
            self.knock_in_scenario = knock_in_scenario & self.not_knock_out



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .bsm_process import *
from .heston_process import *
from .stoch_process import *

__all__ = ['GeneralizedBSMProcess', 'HestonProcess', 'StochProcessBase']



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .accumulator import Accumulator
from .range_accural import RangeAccural



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from scipy.optimize import leastsq
from pricelib.common.utilities.enums import CallPut, ExerciseType
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import McEngine


# pylint: disable=invalid-name
def quadratic_func(x, p):
    """二次函数"""
    a, b, c = p
    return a * (x ** 2) + b * x + c


# pylint: disable=invalid-name
def residuals_func(p, y, x):
    """残差函数"""
    return y - quadratic_func(x, p)


class MCVanillaEngine(McEngine):
    """香草期权 Monte Carlo 模拟定价引擎
        支持欧式期权和美式期权"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        tau = (prod.end_date - calculate_date).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        n_step = int(tau * prod.t_step_per_year)
        paths = self.path_generator(n_step=n_step, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()
        r = self.process.interest(tau)
        if prod.exercise_type == ExerciseType.European:
            price = np.mean(np.maximum(prod.callput.value * (paths[-1, :] - prod.strike), 0)) * np.exp(
                -r * tau)
            return price
        if prod.exercise_type == ExerciseType.American:  # 美式期权，LSMC方法
            v_grid = np.zeros(shape=paths.shape)
            v_grid[-1, :] = np.maximum(prod.callput.value * (paths[-1, :] - prod.strike), 0)
            for i in range(n_step - 1, 0, -1):
                v_grid[i, :] = v_grid[i + 1, :] * np.exp(-r * 1 / prod.t_step_per_year)
                if prod.callput == CallPut.Put:  # 认沽
                    lsm_id = np.where(paths[i, :] <= prod.strike)
                else:  # prod.callput == CallPut.Call: 认购
                    lsm_id = np.where(paths[i, :] >= prod.strike)
                if np.array(lsm_id).size <= 2:  # 拟合曲线有三个参数，若实值点<3则无法拟合
                    continue
                xdata = np.squeeze(paths[i, lsm_id])
                ydata = np.squeeze(v_grid[i + 1, lsm_id]) * np.exp(-r * 1 / prod.t_step_per_year)
                lsm_params = leastsq(residuals_func, x0=np.array([0.01, -1.0, spot]), args=(ydata, xdata))
                v_grid[i, lsm_id] = np.maximum(quadratic_func(paths[i, lsm_id], lsm_params[0]),
                                               prod.callput.value * (paths[i, lsm_id] - prod.strike))
            return np.mean(v_grid[1, :] * np.exp(-r * 1 / prod.t_step_per_year))
        raise ValueError("不支持的行权方式, 香草mc定价引擎仅支持欧式期权和美式期权")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import CallPut, ExerciseType
from pricelib.common.pricing_engine_base import FdmEngine, FdmGridwithBound, FdmGrid
from pricelib.common.time import global_evaluation_date


class FdmVanillaEngine(FdmEngine):
    """香草期权PDE有限差分法定价引擎
    支持欧式期权和美式期权"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值，PDE有限差分定价, 时间逆向迭代
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        smax = self.n_smax * spot  # 价格网格上界, 默认为4倍初始价格
        # 返回PDE系数组件abc的函数
        fn_pde_coef = self.process.get_fn_pde_coef(maturity, spot)

        # 生成边界条件：在s=0处，call近似为0; 在smax处，put近似为0。根据看涨看跌平价 call-put = S*exp(-qT) - K*exp(-rT),
        # 可得对应的put(0) = K * exp(-rT), call(smax) = smax * exp(-qT) - K * exp(-rT)
        if prod.callput == CallPut.Call:
            fn_callput = [lambda u: 0, lambda u: smax * self.process.div.disc_factor(maturity, u) -
                          prod.strike * self.process.interest.disc_factor(maturity, u)]
        else:  # CallPut.Put
            fn_callput = [lambda u: prod.strike * self.process.interest.disc_factor(maturity, u),
                          lambda u: 0]
        if prod.exercise_type == ExerciseType.European:
            fdm = FdmGridwithBound(smax=smax, t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                                   fn_pde_coef=fn_pde_coef, fdm_theta=self.fdm_theta)
            fdm.set_boundary_condition(fn_callput)
            # 初始化到期日payoff价值yv=max(callput * (F-K), 0)。fdm.xv是log(F)
            yv = np.maximum(prod.callput.value * (fdm.s_vec - prod.strike), 0)
            yv = fdm.evolve_with_interval(start=maturity, end=0, yv=yv)  # 时间逆向迭代求解
        elif prod.exercise_type == ExerciseType.American:
            t_step = round(maturity * prod.t_step_per_year)  # 时间格点数
            t_vec = np.linspace(0, maturity, t_step + 1)  # 时间向量
            dt = (t_vec[-1] - t_vec[0]) / (t_vec.size - 1)  # 时间步长
            fdm = FdmGrid(smax=smax, t_step_per_year=prod.t_step_per_year, s_step=self.s_step, fn_pde_coef=fn_pde_coef,
                          fdm_theta=self.fdm_theta, maturity=maturity)
            fdm.v_grid[0] = fn_callput[0](t_vec)
            fdm.v_grid[-1] = fn_callput[1](t_vec)
            yv = np.maximum(prod.callput.value * (fdm.s_vec - prod.strike), 0)
            v_vec = prod.callput.value * (fdm.s_vec - prod.strike)
            for j in range(t_step - 1, -1, -1):
                yv = fdm.evolve(j, yv, dt)  # 每组(p, q)更新。如果在障碍观察区间内，更新fd_bound；在观察区间外，更新fd_full
                yv = np.maximum(yv, v_vec)
        else:
            raise ValueError("不支持的期权类型，仅支持欧式或美式香草期权")
        return np.float64((fdm.functionize(yv, kind="cubic")(spot)))



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from abc import ABCMeta, abstractmethod  # 引入ABCMeta 和 abstractmethod 来定义抽象类和抽象方法
import numpy as np
from .utility import logging


class HashableArray(np.ndarray):
    """自定义HashableArray类，继承自np.ndarray。自定义这个类的目的是让可哈希的array能够和@functools.lru_cache装饰器配合使用。
    lru_cache实现了 least recently used cache 最近最少使用缓存。它会为函数建立一个哈希表，当函数新的input已经存在于哈希表的键，则直接返回哈希表中保存的对应output值。
    但是np.ndarray是不可哈希对象，无法作为哈希表的键。为了当函数输入的参数是array时，能够使用lru_cache，需要自定义一个可哈希的array类。"""

    def __new__(cls, *args):
        """覆写np.ndarray的实例化方法，在原来实例化方法的基础上，将实例转化为HashableArray对象。
        类的实例化包含两个步骤: (1)在内存中创建对象，即开辟一块内存空间来存放类的实例Instance; (2)初始化对象，即给实例的属性赋予初始值，例如全部填0
        在 python 中，第一步由 __new__ 函数负责，第二步由 __init__ 函数负责。
        np.asarray方法可以将所有的*args参数转换为数组np.ndarray，然后view方法将数组np.ndarray转化为cls类(即HashableArray类)。
        view是"视图", 与np的copy方法相反，view是数据的引用，修改view返回的变量的值，会影响原始数据的值，而修改view维数和类型不会影响原始数据的维数和类型
        """
        return np.asarray(*args).view(cls)

    def __getitem__(self, *args):
        """覆写np.ndarray的切片方法__getitem__。array的索引切片本来返回的是array对象，将其转化为HashableArray对象"""
        return HashableArray(super().__getitem__(*args))

    def __hash__(self):
        """__hash__方法用于实现对象的哈希值计算。当对象被用作字典的键或者放入集合中时，__hash__方法会被调用来计算哈希值。
        这里先用np.ndarray的tostring方法将数组的值转化为bytes，然后用hash方法将bytes转化为一个整数，即哈希值。"""
        return hash(self.tobytes())

    def __eq__(self, other):
        """__eq__自定义类型比较，这是为了与__hash__方法配合，当想要看一个新数组other是否在哈希表中时，会与哈希表已有的作为键的数组进行比较。
        只要np.array_equal两个数组每个位置的元素相同，即认为两个数组相同。那么就可以认为新数组other已经在哈希表中，直接返回哈希表对应的值"""
        return np.array_equal(self, other)


class Observer(metaclass=ABCMeta):  # metaclass=ABCMeta: 使创建的类为抽象类
    """观察者的基类"""

    @abstractmethod  # 子类必须实现抽象类的抽象方法，这样可以使子类有统一的类方法
    def update(self, observable, *args, **kwargs):
        """收到被观察者的通知时，更新观察者。
        例: 自动改变被观察者的行情属性的值，更新缓存结果
        def update(self, observable, *args, **kwargs):
            if observable is self._s0:
                print(f"观察者收到来自被观察者 s0 的通知：值更新为{observable.data}")
            elif observable is self._sigma:
                print(f"观察者收到来自被观察者 sigma 的通知：值更新为{observable.data}")
            self.__cache = self._calculate(self._s0.data, self._sigma.data)
            print(f"缓存值更新为{self.__cache}")"""

    @abstractmethod
    def remove_self(self):
        """删除对象自己，del之前，先将自己从被观察者的列表中移除
        例:
        self._s0.remove_observer(self)
        self._sigma.remove_observer(self)
        del self"""


class Observable(metaclass=ABCMeta):
    """被观察者的基类"""

    def __init__(self):
        """初始化观察者列表"""
        self.__observers = []

    @property
    def observers_list(self):
        return self.__observers

    def add_observer(self, observer):
        """在观察者列表中添加新的观察者"""
        self.__observers.append(observer)

    def remove_observer(self, observer):
        """在观察者列表中删除指定的观察者"""
        if observer in self.__observers:
            self.__observers.remove(observer)

    def notify_observers(self, *args, **kwargs):
        """当被观察者的内容或状态变化时，通知所有的观察者自动更新"""
        for o in self.__observers:
            o.update(self, *args, **kwargs)

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """被观察者一般是市场行情或估值参数，因此有__call__方法，返回指定时间(及标的价格)下的值"""


class SimpleQuote(Observable):
    """简单行情类，被观察者"""

    def __init__(self, value, name=None):
        super().__init__()
        self.__data = value
        self.__name = name

    @property
    def data(self):
        return self.__data

    @data.setter
    def data(self, new_value):
        """覆写给对象的属性赋值的方法。在被观察者的值改变时，自动向观察者发送通知"""
        try:
            self.__data = float(new_value)  # 被观察者值变化
        except ValueError as e:
            logging.error(f"ValueError:{e}")
        else:
            self.notify_observers(new_value, name=self.__name)  # 自动向观察者发送通知

    def __call__(self, *args, **kwargs):
        return self.data

    @property
    def name(self):
        return self.__name


if __name__ == "__main__":
    # 举例: 标的价格S和波动率sigma是被观察者，BsEulerEvolve是观察者，被观察者的值变化，BsEulerEvolve的__cachevalue会自动变化
    # 并且，通过使用@lru_cache(maxsize=1)装饰器，BsEulerEvolve是观察者会自动缓存最近1次的计算结果，
    # 当输入不变时，直接返回上一次的结果；当输入变化时，重新计算__cachevalue并更新保存的计算结果。
    from functools import lru_cache


    class BsEulerEvolve(Observer):
        """BS SDE欧拉离散更新, 观察者"""

        def __init__(self, underlying: SimpleQuote, sigma: SimpleQuote, riskfree: float, dt: float):
            assert (isinstance(underlying, Observable) and
                    isinstance(sigma, Observable)), "Error: underlying和sigma不是Observable实例"
            super().__init__()
            self.__cachevalue = None  # 存储结果
            self._riskfree = riskfree
            self._dt = dt
            self._s0 = underlying
            underlying.add_observer(self)  # 将本实例加入underlying观察者列表
            self._sigma = sigma
            sigma.add_observer(self)  # 将本实例加入sigma观察者列表

        def evolve(self):
            self.__cachevalue = self._euler_discretization(self._s0.data, self._sigma.data)
            logging.info(f"evolve result = {self.__cachevalue}")

        @lru_cache(maxsize=1)  # 只缓存最近的一种结果
        def _euler_discretization(self, spot, sigma):
            return spot * (1 + self._riskfree * self._dt + sigma * np.random.rand() * np.sqrt(self._dt))

        def update(self, observable, *args, **kwargs):
            if isinstance(observable, SimpleQuote):
                if observable is self._s0:
                    logging.info(f"观察者收到来自被观察者s0的通知：数据更新为{observable.data}")
                elif observable is self._sigma:
                    logging.info(f"观察者收到来自被观察者sigma的通知：数据更新为{observable.data}")
                logging.info(f"update传入的位置参数: {args}，关键字参数: {kwargs}")
                logging.info(f"被观察者的类名 = {type(observable).__name__}")
                self.evolve()

        def remove_self(self):
            """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
            self._s0.remove_observer(self)
            self._sigma.remove_observer(self)
            del self


    # 被观察者——标的价格初始值为100
    S = SimpleQuote(value=100, name="标的价格")
    vol = SimpleQuote(value=0.2, name="波动率")
    # 观察者——Black Scholes SDE欧拉离散迭代一步evolve
    dynamic = BsEulerEvolve(underlying=S, sigma=vol, riskfree=0.02, dt=0.005)
    print("1. 初始值:")
    dynamic.evolve()
    print("1'. 再次调用evolve，由于被观察者无变化，直接返回缓存结果，观察者值不变")
    dynamic.evolve()
    print("2. 被观察者【S】改变，观察者evolve自动重新计算:")
    S.data = 105
    print("2'. 再次调用evolve，由于被观察者无变化，直接返回缓存结果，观察者值不变")
    dynamic.evolve()
    print("3. 被观察者【vol】改变，观察者evolve自动重新计算:")
    vol.data = 0.3
    print(f"4. 删除观察者对象，会从被观察者列表中移除自己\nbefore: {S.observers_list, vol.observers_list}")
    dynamic.remove_self()
    print(f"after: {S.observers_list, vol.observers_list}")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from contextlib import suppress
from pricelib.common.utilities.enums import StatusType, CallPut
from pricelib.common.time import (global_evaluation_date, Schedule, CN_CALENDAR, AnnualDays)
from pricelib.common.utilities.patterns import Observer
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.mc_engines import MCAccumulatorEngine


class Accumulator(OptionBase, Observer):
    """累购累沽期权
    以累购为例：行权价通常低于二级市场标的现价，敲出障碍价通常高于二级市场标的现价。
             在每个交易日，如果标的价格在执行价之下，以行权价购买N份标的，N是杠杆比例，例如N=2；
                         如果标的价格在执行价之上，以行权价购买1份标的；
             在敲出观察日，如果标的价格在敲出障碍价之上，发生敲出，合约提前终止，结算收益。
             最终的收益即是累计的所有标的份数 x (结算时的标的价格 - 行权价)。
    累沽与累沽相反，行权价通常高于二级市场标的现价，敲出障碍价通常低于二级市场标的现价。
    """

    def __init__(self, s0, barrier_out=None, strike=None, leverage_ratio=None, obs_dates=None, margin_lvl=0.2,
                 engine=None, status=StatusType.NoTouch, maturity=None, start_date=None, end_date=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        产品参数:
            s0: float，标的初始价格
            barrier_out: float，敲出障碍价
            strike: float，行权价
            leverage_ratio: float, 杠杆比例，累购期权下，当标的价格在执行价之下时，购买leverage_ratio份标的；
                                                    当标的价格在执行价之上、敲出价之下时，购买1份标的
            margin_lvl: float，保证金比例，对于定价没有影响
            status: 敲出状态，默认为StatusType.NoTouch表示未敲出
            engine: 定价引擎，PricingEngine类，仅支持Monte Carlo模拟 MCAccumulatorEngine
        时间参数: 要么输入年化期限，要么输入起始日和到期日；敲出观察日可缺省
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            obs_dates: List[datetime.date]，敲出观察日，可缺省，缺省时会自动生成每月观察的敲出日期序列(已根据节假日调整)
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用蒙特卡洛模拟定价引擎 (BSM模型)
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.s0 = s0  # 标的初始价格
        self.trade_calendar = trade_calendar  # 交易日历
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        if maturity is None:
            self.maturity = (end_date - start_date).days / annual_days.value
        else:
            self.maturity = maturity
        self.barrier_out = barrier_out  # 敲出线
        self.strike = strike  # 行权价
        if barrier_out > strike:  # 累购
            self.callput = CallPut.Call
        else:  # 累沽
            self.callput = CallPut.Put
        if obs_dates is None:  # 观察日列表
            self.obs_dates = Schedule(trade_calendar=trade_calendar, start=self.start_date, end=self.end_date,
                                      freq='m', lock_term=0)  # 累购累沽无锁定期
            self.end_date = self.obs_dates.end  # 修正结束时间
        else:
            self.obs_dates = Schedule(trade_calendar=trade_calendar, date_schedule=obs_dates)
        self.leverage_ratio = leverage_ratio
        self.status = status
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            self.engine = MCAccumulatorEngine(s=s, r=r, q=q, vol=vol)


    def set_pricing_engine(self, engine):
        """切换定价引擎"""
        with suppress(AttributeError, ValueError):
            self.engine.process.spot.remove_observer(self)
        self.engine = engine
        self.engine.process.spot.add_observer(self)
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.engine.process.spot.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """收到标的价格变化的通知时，自动更新是否已经敲出"""
        if observable == self.engine.process.spot:
            if self.callput == CallPut.Call and observable.value >= self.barrier_out:
                self.status = StatusType.UpTouch
            elif self.callput == CallPut.Put and observable.value <= self.barrier_out:
                self.status = StatusType.DownTouch

    def __repr__(self):
        """返回期权的描述"""
        if self.barrier_out >= self.strike:
            return "累购(Accumulator)"
        return "累沽(Decumulator)"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price
